"""
Apple Objective-C Runtime Support

This module implements support for analyzing the Apple Objective-C runtime. As
expected, Apple provides no oficial specification for the internal ABI of ObjC
and no guarantees of its stability, and so this module is not guaranteed to
work on all versions of Darwin.
"""

from __future__ import annotations

from collections.abc import Callable
from collections.abc import Generator
from typing import Generic
from typing import TypeVar

from typing_extensions import override

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.macho
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.lib.config
import pwndbg.lib.functions
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.dbg_mod import Type

T = TypeVar("T")


class _IdRaw:
    """
    Pointer to an Objective-C object in the heap.
    """

    def __init__(self, ptr: int):
        self.addr = ptr


class _IdTagged:
    """
    Tagged pointer to an Objective-C object.

    This is a bit of a misnomer, as tagged pointers may not be pointers at all,
    and the data for the entire object may be contained in the payload, with no
    backing allocation in the heap. It is up to the class to determine how to
    decode the payload properly.
    """

    def __init__(self, tag: int, payload: int, extended: bool):
        self.tag = tag
        self.payload = payload
        self.extended = extended

    def lookup_class(self) -> Class:
        """
        Looks up the class object matching the tag in this pointer.
        """
        classes = _tagged_pointer_classes()

        if self.extended:
            classes += self.tag - 256
        else:
            classes += self.tag

        ptr = pwndbg.aglib.memory.read_pointer_width(int(classes.address))
        ptr = _ptrauth_strip(ptr)

        return Class(ptr)


class _IsaPtr:
    """
    Pointer to an `isa_t` structure.
    """

    ISA_MASK = 0x0000000FFFFFFFF8
    "Mask of bits containing just the authenticated class pointer."

    def __init__(self, addr: int):
        self._addr = addr

    def _read(self) -> int:
        """
        Read the bits of the `isa_t` structure into an integer.
        """
        return pwndbg.aglib.memory.read_pointer_width(self._addr)

    def get_class(self) -> Class:
        ptr = self._read() & _IsaPtr.ISA_MASK
        ptr = _ptrauth_strip(ptr)

        return Class(ptr)


def _isa_class_mask() -> int:
    return pwndbg.aglib.memory.read_pointer_width(
        pwndbg.aglib.symbol.lookup_symbol_addr("objc_debug_isa_class_mask")
    )


class _ClassRoPtr:
    RO_META = 0x1
    RO_ROOT = 0x2
    RO_HAS_CXX_STRUCTORS = 0x4
    RO_HIDDEN = 0x10
    RO_EXCEPTION = 0x20
    RO_HAS_SWIFT_INITIALIZER = 0x40
    RO_IS_ARC = 0x80
    RO_HAS_CXX_DTOR_ONLY = 0x100
    RO_HAS_WEAK_WITHOUT_ARC = 0x200
    RO_FORBIDS_ASSOCIATED_OBJECTS = 0x400
    RO_FROM_BUNDLE = 0x20000000
    RO_FUTURE = 0x40000000
    RO_REALIZED = 0x80000000

    def __init__(self, addr: int):
        self._ptr = addr

    def name(self) -> bytes:
        ptr = pwndbg.aglib.memory.read_pointer_width(self._ptr + 24)
        return pwndbg.aglib.memory.string(ptr)

    def flags(self) -> int:
        return pwndbg.aglib.memory.u32(self._ptr)

    def methods(self) -> Generator[Method]:
        ptr = pwndbg.aglib.memory.read_pointer_width(self._ptr + 32)

        if ptr & 1 == 0:
            if ptr == 0:
                return
            yield from _MethodList(ptr).entries()
        else:
            if ptr & ~1 == 0:
                # Not expected to happen, but better safe than sorry.
                return

            list_of_lists = _RelativeListOfLists(_MethodList, ptr & ~1)
            for lst in list_of_lists.entries():
                if lst is None:
                    continue
                yield from lst.get_list().entries()

    def ivars(self) -> Generator[InstanceVariable]:
        ptr = pwndbg.aglib.memory.read_pointer_width(self._ptr + 0x30)
        if ptr != 0:
            yield from _IVarList(ptr).entries()

    def properties(self) -> Generator[ClassProperty]:
        ptr = pwndbg.aglib.memory.read_pointer_width(self._ptr + 0x40)
        if ptr & 1 == 0:
            if ptr == 0:
                return

            yield from _ClassPropertyList(ptr).entries()
        else:
            if ptr & ~1 == 0:
                # Not expected to happen, but better safe than sorry.
                return

            list_of_lists = _RelativeListOfLists(_ClassPropertyList, ptr & ~1)
            for lst in list_of_lists.entries():
                if lst is None:
                    continue
                yield from lst.get_list().entries()


class _ClassRwExtPtr:
    def __init__(self, ptr: int):
        self._ptr = ptr

    def ro(self) -> _ClassRoPtr:
        ptr = pwndbg.aglib.memory.read_pointer_width(self._ptr)
        ptr = _ptrauth_strip(ptr)
        return _ClassRoPtr(ptr)

    def methods(self) -> _ListArray[Method]:
        return _ListArray(_MethodList, self._ptr + pwndbg.aglib.typeinfo.ptrsize)

    def properties(self) -> _ListArray[ClassProperty]:
        return _ListArray(_ClassPropertyList, self._ptr + 2 * pwndbg.aglib.typeinfo.ptrsize)

    def demangled_name(self) -> bytes | None:
        ptr = pwndbg.aglib.memory.read_pointer_width(self._ptr + 4 * pwndbg.aglib.typeinfo.ptrsize)
        if ptr == 0:
            return None
        return pwndbg.aglib.memory.string(ptr)

    def version(self) -> int:
        return pwndbg.aglib.memory.u32(self._ptr + 5 * pwndbg.aglib.typeinfo.ptrsize)


class _ClassRwPtr:
    RW_REALIZED = 1 << 31

    def __init__(self, ptr: int):
        self._ptr = ptr

    def ro_or_rw_ext(self) -> _ClassRoPtr | _ClassRwExtPtr:
        ptr = pwndbg.aglib.memory.read_pointer_width(self._ptr + 8)
        if ptr & 1 == 1:
            return _ClassRwExtPtr(ptr & ~1)
        return _ClassRoPtr(ptr)


class _ClassDataBitsPtr:
    """
    Pointer to a `class_data_bits_t` structure.
    """

    FAST_IS_RW_POINTER = 0x8000000000000000
    FAST_IS_SWIFT_LEGACY = 0x1
    FAST_IS_SWIFT_STABLE = 0x2
    FAST_HAS_DEFAULT_RR = 0x4

    FAST_DATA_MASK = 0x0F007FFFFFFFFFF8

    def __init__(self, ptr: int):
        self._ptr = ptr

    def data(self) -> _ClassRoPtr | _ClassRwPtr:
        if self._is_rw():
            return _ClassRwPtr(self._data_addr())

        return _ClassRoPtr(self._data_addr())

    def _is_rw(self) -> bool:
        return ((self._ptr & _ClassDataBitsPtr.FAST_IS_RW_POINTER) != 0) or (
            (self._flags() & _ClassRwPtr.RW_REALIZED) != 0
        )

    def _data_addr(self) -> int:
        return _ptrauth_strip(self._ptr) & _ClassDataBitsPtr.FAST_DATA_MASK

    def _flags(self) -> int:
        return pwndbg.aglib.memory.u32(self._data_addr())


class _EntList(Generic[T]):
    """
    Entity list.
    """

    _flags_mask: int = 0
    "Mask for the flag bits of `entsizeAndFlags`"

    def __init__(self, ptr: int):
        self._addr = self._addr_from_ptr(ptr)
        self._ptr = ptr

    def _entsize_and_flags(self) -> int:
        return pwndbg.aglib.memory.u32(self._addr)

    def _entries(self) -> int:
        return pwndbg.aglib.memory.u32(self._addr + 4)

    def flags(self) -> int:
        return self._entsize_and_flags() & self._flags_mask

    def entsize(self) -> int:
        return self._entsize_and_flags() & ~self._flags_mask

    def _modify_pointer(self, ptr: int) -> int:
        return ptr

    def _from_ptr(self, ptr: int) -> T:
        """
        Build the type of this list from a pointer.

        Must be implemented by the specialized class.
        """
        raise NotImplementedError()

    def _addr_from_ptr(self, ptr: int) -> int:
        """
        Strip any metadata from the pointer to this list.

        Must be implemented by the specialized class.
        """
        raise NotImplementedError()

    def __len__(self) -> int:
        return self._entries()

    def get(self, i: int) -> T:
        if i >= len(self):
            raise IndexError(f"Index {i} is out-of-range for entlist with {len(self)} entries")

        return self._from_ptr(self._modify_pointer(self._addr + 8 + i * self.entsize()))

    def entries(self) -> Generator[T]:
        for i in range(len(self)):
            yield self.get(i)


class _RelativeListOfListsEntry(Generic[T]):
    def __init__(self, ty: Callable[[int], _EntList[T]], ptr: int):
        self._ptr = ptr
        self._ty = ty

    def image_index(self) -> int:
        return pwndbg.aglib.memory.u64(self._ptr) & 0xFFFF

    def _list_offset(self) -> int:
        return pwndbg.aglib.memory.s64(self._ptr) >> 16

    def get_list(self) -> _EntList[T]:
        return self._ty(self._ptr + self._list_offset())


class _RelativeListOfLists(
    _EntList[_RelativeListOfListsEntry[T] | None],
    Generic[T],
):
    """
    An array of relative pointers to lists.

    This corresponds to the `relative_list_list_t` type in libobjc.
    """

    def __init__(self, ty: Callable[[int], _EntList[T]], ptr: int):
        super().__init__(ptr)
        self._ty = ty

    @override
    def _addr_from_ptr(self, ptr: int) -> int:
        # Top-Byte-Ignore is assumed for method lists, but method list pointers
        # may have metadata attached to them.
        return ptr & ~0xFF00000000000000

    @override
    def _from_ptr(self, ptr: int) -> _RelativeListOfListsEntry[T] | None:
        entry = _RelativeListOfListsEntry(self._ty, ptr)
        if not _header_info_rw_is_image_loaded(entry.image_index()):
            # The entry is only valid if its corresponding image has been marked
            # as loaded in `objc_debug_headerInfoRWs`.
            return None

        return entry


class _ListArray(Generic[T]):
    """
    A runtime-polymorphic array type for lists. May be a pointer to a list type,
    an array of pointers, or a _RelativeListOfLists, distinguished by a tag in
    a pointer.

    Strangely for Apple, the tagged pointer to the final list is contained
    inside the list array structure, rather than having the whole structure be
    inlined into a pointer value. Suspiciously sane.

    This corresponds to the `list_array_tt` type in libobjc.
    """

    def __init__(self, ty: Callable[[int], _EntList[T]], ptr: int):
        self._ptr = ptr
        self._ty = ty

    def entries(self) -> Generator[T]:
        raw_ptr = pwndbg.aglib.memory.read_pointer_width(self._ptr)

        tag = raw_ptr & 3
        ptr = raw_ptr & ~3

        if ptr == 0:
            return

        if tag == 0:
            # This is just a pointer to the list.
            yield from self._ty(ptr).entries()
        elif tag == 1:
            # This is an array of lists.
            count = pwndbg.aglib.memory.u32(ptr)
            for i in range(count):
                yield from self._ty(
                    pwndbg.aglib.memory.read_pointer_width(
                        ptr + 8 + i * pwndbg.aglib.typeinfo.ptrsize
                    )
                ).entries()
        elif tag == 2:
            # This is a relative list of lists.
            for ll in _RelativeListOfLists(self._ty, ptr).entries():
                yield from ll.get_list().entries()


def _header_info_rw_is_image_loaded(index: int) -> bool:
    """
    Queries `objc_debug_headerInfoRWs` and checks whether the image with the
    given index is loaded.
    """
    addr = pwndbg.aglib.memory.read_pointer_width(
        pwndbg.aglib.symbol.lookup_symbol_addr("objc_debug_headerInfoRWs")
    )

    count = pwndbg.aglib.memory.u32(addr)
    entsize = pwndbg.aglib.memory.u32(addr + 4)

    if index >= count:
        raise IndexError(
            f"Image index {index} is out-of-bounds for headerInfoRWs structure with {count} entries"
        )

    return pwndbg.aglib.memory.read_pointer_width(addr + 8 + entsize * index) & 1 == 1


def _tagged_pointer_classes() -> pwndbg.dbg_mod.Value:
    """
    The Objective-C runtime tagged pointer classs list.

    The classes to which the tag values in a tagged pointer corresponds are not
    fixed, and are instead stored in a runtime-global array that gets looked up
    when a message is sent.
    """
    return pwndbg.aglib.symbol.lookup_symbol("objc_debug_taggedpointer_classes").cast(
        pwndbg.aglib.typeinfo.void.pointer().pointer()
    )


def _ptr_obfuscation_value() -> int:
    """
    The Objective-C runtime obfuscates tagged pointer values.
    """
    return pwndbg.aglib.memory.read_pointer_width(
        pwndbg.aglib.symbol.lookup_symbol_addr("objc_debug_taggedpointer_obfuscator")
    )


def _try_decode_tagged_split(ptr: int) -> _IdTagged | None:
    """
    Decodes a tagged pointer encoded in the split-tag scheme, if it is tagged.

    This is the encoding scheme used in modern - iOS 14 and newer - ARM64
    platforms.

    If the pointer is not tagged, returns `None`.
    """
    if ptr & 0x8000000000000000 == 0:
        # Not a tagged pointer.
        return None

    if ptr & 7 == 7:
        # This is an extended tag with a 52-bit payload.
        tag = (ptr >> 55) & 0xFF
        payload = (ptr >> 3) & 0xFFFFFFFFFFFFF
        extended = True
    else:
        # This is a short tag with a 60-bit payload.
        tag = ptr & 7
        payload = (ptr >> 3) & 0xFFFFFFFFFFFFFFF
        extended = False

    return _IdTagged(tag, payload, extended)


def _try_decode_tagged_lsb(ptr: int) -> _IdTagged:
    """
    Decodes a tagged pointer encoded in the LSB-tag scheme, if it is tagged.

    This is the encoding scheme used in all x86-64 versions of Darwin.

    If the pointer is not tagged, returns `None`.
    """
    if ptr & 1 == 0:
        # Not a tagged pointer.
        return None

    if ptr & 14 == 14:
        # This is an extended tag with a 52-bit payload.
        tag = (ptr >> 4) & 0xFF
        payload = ptr >> 12
        extended = True
    else:
        # This is a short tage with a 60-bit payload.
        tag = (ptr >> 1) & 7
        payload = ptr >> 4
        extended = False

    return _IdTagged(tag, payload, extended)


def _decode_prog_id(ptr: int) -> _IdRaw | _IdTagged:
    """
    Given an Objective-C program, decode it.
    """

    # First, check for tagged pointers.
    tagged = None
    match pwndbg.aglib.arch.name:
        case "aarch64":
            tagged = _try_decode_tagged_split(ptr)
        case "x86-64":
            tagged = _try_decode_tagged_lsb(ptr)
        case other:
            raise AssertionError(f"Unexpected Objective-C architecture: {other}")
    if tagged is not None:
        # Successfuly decoded the tagged pointer.
        return tagged

    # This is a direct pointer.
    return _IdRaw(ptr)


def _ptrauth_strip(ptr: int) -> int:
    """
    Strip pointer signing information from a given signed pointer.
    """
    return ptr & 0xFFFFFFFFFFFF


class Object:
    _addr: int
    "Object pointer value, as seen in the program. May be tagged, obfuscated, authenticated."

    _id: _IdRaw | _IdTagged
    "Decoded object pointer value. May be tagged."

    def __init__(self, addr: int):
        self._addr = addr
        self._id = _decode_prog_id(addr)

    @property
    def cls(self) -> Class | None:
        if isinstance(self._id, _IdRaw):
            isa = _IsaPtr(self._id.addr)
            return isa.get_class()
        if isinstance(self._id, _IdTagged):
            return self._id.lookup_class()


class Class(Object):
    def __init__(self, addr: int):
        super().__init__(addr)
        assert isinstance(self._id, _IdRaw), "Class pointers are never tagged"

    def _data_bits(self) -> _ClassDataBitsPtr:
        # MyPy fails if we don't check this a second time.
        assert isinstance(self._id, _IdRaw), "Class pointers are never tagged"
        ptr = pwndbg.aglib.memory.read_pointer_width(self._id.addr + 32)
        ptr = _ptrauth_strip(ptr)
        return _ClassDataBitsPtr(ptr)

    def _ro(self) -> _ClassRoPtr:
        data = self._data_bits().data()
        if isinstance(data, _ClassRoPtr):
            return data
        if isinstance(data, _ClassRwPtr):
            ro_or_rw_ext = data.ro_or_rw_ext()
            if isinstance(ro_or_rw_ext, _ClassRwExtPtr):
                return ro_or_rw_ext.ro()
            if isinstance(ro_or_rw_ext, _ClassRoPtr):
                return ro_or_rw_ext
            # FIXME: Should be `typing.assert_never`, needs Python 3.11
            assert False
        else:
            # FIXME: Should be `typing.assert_never`, needs Python 3.11
            assert False

    def _rw_ext(self) -> _ClassRwExtPtr | None:
        data = self._data_bits().data()
        if isinstance(data, _ClassRoPtr):
            return None
        if isinstance(data, _ClassRwPtr):
            ro_or_rw_ext = data.ro_or_rw_ext()
            if isinstance(ro_or_rw_ext, _ClassRwExtPtr):
                return ro_or_rw_ext
            if isinstance(ro_or_rw_ext, _ClassRoPtr):
                return None
            # FIXME: Should be `typing.assert_never`, needs Python 3.11
            assert False
        else:
            # FIXME: Should be `typing.assert_never`, needs Python 3.11
            assert False

    @property
    def superclass(self) -> Class | None:
        # MyPy fails if we don't check this a second time.
        assert isinstance(self._id, _IdRaw), "Class pointers are never tagged"

        if self._ro().flags() & _ClassRoPtr.RO_ROOT != 0:
            # This is a root class, and thus has no superclass.
            return None

        ptr_addr = self._id.addr + pwndbg.aglib.typeinfo.ptrsize
        ptr = pwndbg.aglib.memory.read_pointer_width(ptr_addr)
        ptr = _ptrauth_strip(ptr)

        return Class(ptr)

    @property
    def name(self) -> bytes:
        return self._ro().name()

    @property
    def methods(self) -> Generator[Method]:
        if (rw_ext := self._rw_ext()) is not None:
            # Return the methods added to the class at runtime from the Class
            # R/W structure, which also include the base methods.
            yield from rw_ext.methods().entries()
        else:
            # Return the base methods.
            yield from self._ro().methods()

    @property
    def ivars(self) -> Generator[InstanceVariable]:
        yield from self._ro().ivars()

    @property
    def properties(self) -> Generator[ClassProperty]:
        if (rw_ext := self._rw_ext()) is not None:
            # Return the properties added to the class at runtime from the Class
            # R/W structure, which also include the base properties.
            yield from rw_ext.properties().entries()
        else:
            # Return the base properties.
            yield from self._ro().properties()

    @property
    def is_metaclass(self) -> bool:
        return (self._ro().flags() & _ClassRoPtr.RO_META) != 0

    @override
    @property
    def cls(self) -> Class | None:
        if self.is_metaclass:
            # Following this pointer in metaclasses is weird. Users are better
            # served following the superclass chain, instead.
            return None
        return super().cls


class InstanceVariable:
    """
    An Objective-C Instance Variable.

    Instance Variables are NOT objects!
    """

    def __init__(self, ptr: int):
        self._ptr = ptr

    @property
    def offset(self) -> int:
        """
        The offset in bytes of this value from the start of the object instance.
        """
        return pwndbg.aglib.memory.s32(pwndbg.aglib.memory.read_pointer_width(self._ptr))

    @property
    def name(self) -> bytes:
        """
        The name of this instance variable.
        """
        return pwndbg.aglib.memory.string(
            pwndbg.aglib.memory.read_pointer_width(self._ptr + pwndbg.aglib.typeinfo.ptrsize)
        )

    @property
    def typename(self) -> bytes:
        """
        The name of the type of this instance variable.
        """
        return pwndbg.aglib.memory.string(
            pwndbg.aglib.memory.read_pointer_width(self._ptr + pwndbg.aglib.typeinfo.ptrsize * 2)
        )

    @property
    def alignment(self) -> int:
        """
        The alignment of this instance variable, in bytes.
        """
        align_log2 = pwndbg.aglib.memory.u32(self._ptr + pwndbg.aglib.typeinfo.ptrsize * 3)

        # All ones indicates the natural alignment of a pointer.
        if align_log2 == 0xFFFFFFFF:
            return pwndbg.aglib.typeinfo.ptrsize

        return 1 << align_log2

    @property
    def size(self) -> int:
        """
        The size of this instance variable, in bytes.
        """
        return pwndbg.aglib.memory.u32(self._ptr + pwndbg.aglib.typeinfo.ptrsize * 3 + 4)


class ClassProperty:
    def __init__(self, ptr: int):
        self._ptr = ptr

    @property
    def name(self) -> bytes:
        """
        The name of this class property.
        """
        return pwndbg.aglib.memory.string(pwndbg.aglib.memory.read_pointer_width(self._ptr))

    @property
    def value(self) -> bytes:
        """
        The value of this property.
        """
        return pwndbg.aglib.memory.string(
            pwndbg.aglib.memory.read_pointer_width(self._ptr + pwndbg.aglib.typeinfo.ptrsize)
        )


class Selector:
    """
    An Objective-C Selector.

    Selectors are NOT objects!
    """

    def __init__(self, ptr: int):
        self._ptr = ptr

    @property
    def name(self) -> bytes:
        """
        Retrieves the name of this selector.
        """

        # In the Apple Objective-C runtime, selectors are human-readable strings
        # with unique identities[1]. The identity is simply the pointer to the
        # string itself, guaranteed by the tooling to be unique. To read the
        # name of the selector, then, we can simply follow its identity pointer.
        #
        # [1]: https://web.archive.org/web/20161010081824/http://unixjunkie.blogspot.com/2006/02/nil-and-nil.html
        return pwndbg.aglib.memory.string(self._ptr)


class Method:
    """
    An Objective-C Method Pointer.

    Methods are NOT objects!

    A method pointer can be one of three types: Small, small direct, and big.

    Pointer types are distinguished by the two least significant bits in the
    integer representation of the pointer. A value of `1` is used for both small
    pointer types, while all other values are used to distinguish between the
    signing nuances of big pointers.

    Small pointers 32-bit wide and relative to a given base value. Big pointers
    contain the pointers themselves, and they may or may not be signed.

    Small direct pointers are small pointers that reside in the shared cache, and
    their selectors are relative to @selector(🤯), while the selectors of regular
    small pointers are relative to the pointers themselves.
    """

    def __init__(self, ptr: int):
        self._ptr = ptr

    @property
    def sel(self) -> Selector:
        "The selector this method responds to."
        kind = self._ptr & 3
        base = self._ptr & ~3
        if kind == 1:
            if pwndbg.aglib.macho.shared_cache().is_address_in_shared_cache(base):
                # To resolve selectors of small method pointers in the shared cache,
                # we have to look up the identity of @selector(🤯).
                rel = (
                    pwndbg.aglib.macho.shared_cache().objc_builtin_selectors().lookup("🤯".encode())
                )
                ptr = rel + pwndbg.aglib.memory.s32(base)
            else:
                offset = pwndbg.aglib.memory.s32(base)
                ref = base + offset

                # Non-shared cache values are pointers to selectors.
                ptr = pwndbg.aglib.memory.read_pointer_width(ref)

            return Selector(ptr)
        return Selector(_ptrauth_strip(pwndbg.aglib.memory.read_pointer_width(base)))

    @property
    def types(self) -> bytes:
        "The types of the arguments to this method."
        kind = self._ptr & 3
        base = self._ptr & ~3
        if kind == 1:
            ptr = base + 4
            offset = pwndbg.aglib.memory.s32(ptr)
            addr = ptr + offset
        else:
            ptr = base + 8
            addr = _ptrauth_strip(
                pwndbg.aglib.memory.read_pointer_width(base + pwndbg.aglib.typeinfo.ptrsize)
            )

        return pwndbg.aglib.memory.string(addr)

    @property
    def imp(self) -> int:
        "The pointer to the function that implements this method."
        kind = self._ptr & 3
        base = self._ptr & ~3
        if kind == 1:
            # There's a bit of nuance here.
            #
            # Method swizzling for small pointers is implemented using a global
            # hash map of method pointers to implementation pointers. When
            # getting the IMP pointer for a small pointer, the runtime will
            # first check the global hash map to see if the method has been
            # swizzled, and return the swizzled method if it has. The runtime
            # will do what we do here if method has not been swizzled.
            #
            # Currently, we have no good way to query this map, and no other way
            # to detect that a method has been swizzled, so swizzles to small
            # pointers are unfortunately compeltely invisible to us.
            #
            # TODO: Handle method swizzles for small-pointer-type Objective-C methods.
            ptr = base + 8
            offset = pwndbg.aglib.memory.s32(ptr)
            return ptr + offset
        ptr = base + 16
        return _ptrauth_strip(
            pwndbg.aglib.memory.read_pointer_width(base + pwndbg.aglib.typeinfo.ptrsize)
        )


class _MethodList(_EntList[Method]):
    """
    Method entity list.
    """

    _flags_mask = 0xFFFF0003

    SMALL_METHOD_LIST_FLAG = 0x80000000
    "Indicates that the pointers in this list are small method pointers."

    BIG_SIGNED_METHOD_LIST_FLAG = 0x8000000000000000
    """
    Indicates that the pointers in this list are big and signed.

    Stored as part of the pointer to the method list, rather than in the flags
    field, as is the case with other flags.
    """

    @override
    def _modify_pointer(self, ptr: int) -> int:
        if self.flags() & self.SMALL_METHOD_LIST_FLAG != 0:
            # This is a small pointer list.
            return (ptr & ~3) | 1
        if self._ptr & self.BIG_SIGNED_METHOD_LIST_FLAG:
            # This is a big signed poitner list.
            return (ptr & ~3) | 2
        # No tag or flag. This is a big pointer list.
        return ptr & ~3

    @override
    def _addr_from_ptr(self, ptr: int) -> int:
        # Top-Byte-Ignore is assumed for method lists, but method list pointers
        # may have metadata attached to them.
        return ptr & ~0xFF00000000000000

    @override
    def _from_ptr(self, ptr: int) -> Method:
        return Method(ptr)


class _IVarList(_EntList[InstanceVariable]):
    "IVar entity list."

    _flags_mask = 0

    @override
    def _modify_pointer(self, ptr: int) -> int:
        return ptr

    @override
    def _addr_from_ptr(self, ptr: int) -> int:
        return ptr

    @override
    def _from_ptr(self, ptr: int) -> InstanceVariable:
        return InstanceVariable(ptr)


class _ClassPropertyList(_EntList[ClassProperty]):
    "Class property entity list."

    _flags_mask = 0

    @override
    def _modify_pointer(self, ptr: int) -> int:
        return ptr

    @override
    def _addr_from_ptr(self, ptr: int) -> int:
        return ptr

    @override
    def _from_ptr(self, ptr: int) -> ClassProperty:
        return ClassProperty(ptr)


def _parse_method_type_array(ty: bytes, depth: int) -> tuple[Type, int] | None:
    """
    Parses a typed array entry in an Objective-C method type string.
    """
    if ty[0] != b"[":
        return None
    if (end := ty.find(b"]")) == -1:
        return None
    if (inner := _parse_method_type(ty[1:end], depth + 1)) is None:
        return None

    # Treat arrays as pointers.
    return inner[0].pointer(), inner[1] + 2


def _parse_method_type_pointer(ty: bytes, depth: int) -> tuple[Type, int] | None:
    """
    Parses a typed pointer entry in an Objective-C method type string.
    """
    if ty[0] != b"^":
        return None
    if (inner := _parse_method_type(ty[1:], depth + 1)) is None:
        return None

    return inner[0].pointer(), inner[1] + 1


def _parse_method_type_id_typed(ty: bytes, depth: int) -> tuple[Type, int] | None:
    """
    Parses a typed `id` entry in an Objective-C method type string.
    """
    if ty[:2] != b'@"':
        return None
    if (end := ty.find(b'"', 2)) == -1:
        return None

    # Resolve to `id`, even if we could technically be more specific.
    return pwndbg.aglib.typeinfo.lookup_types("id"), end + 1


def _parse_method_type(ty: bytes, depth: int) -> tuple[Type, int] | None:
    """
    Parses a single entry in an Objective-C method type string.
    """
    if depth > max_method_type_depth.value:
        # Too deep. Reject this type.
        return None

    while ty[0] >= 0x30 and ty[0] < 0x3A:
        # Ignore integers that we don't recognize.
        ty = ty[1:]

    # Try to parse composite types.
    if (res := _parse_method_type_array(ty, depth)) is not None:
        return res
    if (res := _parse_method_type_pointer(ty, depth)) is not None:
        return res
    if (res := _parse_method_type_id_typed(ty, depth)) is not None:
        return res

    # Try to parse atomic types.
    match ty[0:1]:
        case b"v":
            return pwndbg.aglib.typeinfo.void, 1
        case b"f":
            return pwndbg.aglib.typeinfo.lookup_types("float"), 1
        case b"d":
            return pwndbg.aglib.typeinfo.lookup_types("double"), 1
        case b"B":
            return pwndbg.aglib.typeinfo.int32, 1
        case b"*":
            return pwndbg.aglib.typeinfo.char.pointer(), 1
        case b"@":
            return pwndbg.aglib.typeinfo.lookup_types("id"), 1
        case b"#":
            return pwndbg.aglib.typeinfo.lookup_types("Class"), 1
        case b":":
            return pwndbg.aglib.typeinfo.lookup_types("SEL"), 1
        case b"c":
            return pwndbg.aglib.typeinfo.char, 1
        case b"s":
            return pwndbg.aglib.typeinfo.int16, 1
        case b"i":
            return pwndbg.aglib.typeinfo.int32, 1
        case b"l":
            return pwndbg.aglib.typeinfo.int32, 1
        case b"q":
            return pwndbg.aglib.typeinfo.int64, 1
        case b"C":
            return pwndbg.aglib.typeinfo.uchar, 1
        case b"S":
            return pwndbg.aglib.typeinfo.uint16, 1
        case b"I":
            return pwndbg.aglib.typeinfo.uint32, 1
        case b"L":
            return pwndbg.aglib.typeinfo.uint32, 1
        case b"Q":
            return pwndbg.aglib.typeinfo.uint64, 1

    return None


def _parse_method_type_string(ty: bytes) -> Generator[Type | None]:
    """
    Return a generator that yields the type names of the arguments to a method,
    if they can be resolved, yielding `None` for types that we fail to resolve.
    """
    cursor = 0
    yielded = 0
    while yielded < max_method_argument_count.value + 1:
        while cursor < len(ty) and ty[cursor] >= 0x30 and ty[cursor] < 0x3A:
            # Ignore integers.
            cursor += 1

        if cursor >= len(ty):
            break

        res = _parse_method_type(ty[cursor:], 0)

        if res is None:
            yield None
            break

        yield res[0]
        cursor += res[1]
        yielded += 1


max_method_argument_count = pwndbg.config.add_param(
    "objc-max-function-arguments",
    32,
    "maximum number of arguments to resolve for an Objective-C method call",
    param_class=pwndbg.lib.config.PARAM_ZUINTEGER,
)

max_method_type_depth = pwndbg.config.add_param(
    "objc-max-function-types-depth",
    32,
    "maximum allowed depth for a type in an Objective-C method call",
    param_class=pwndbg.lib.config.PARAM_ZUINTEGER,
)


def try_resolve_call_at_current_pc(insn: PwndbgInstruction) -> pwndbg.lib.functions.Function | None:
    """
    Tries to resolve a call to an Objective-C method for an instruction in the
    current Program Counter.
    """
    if not insn.call_like:
        # No point in trying to resolve something that isn't a call.
        return None

    target = pwndbg.aglib.symbol.resolve_addr(insn.target)
    if target == "objc_msgSend":
        # Resolve msgSend.
        #
        # First, try to work out the method implementation from the selector in
        # the current architecture.

        if pwndbg.aglib.arch.name == "aarch64":
            obj_reg = "x0"
            sel_reg = "x1"
        else:
            # Not supported.
            # TODO: Support resolution of Objective-C method calls in x86-64.
            return None

        obj_ptr = pwndbg.aglib.regs.read_reg(obj_reg)
        sel_ptr = pwndbg.aglib.regs.read_reg(sel_reg)

        obj = Object(obj_ptr)
        sel = Selector(sel_ptr)

        # Walk up the class chain and try to find the method value.
        method = None
        clss = obj.cls
        while clss is not None:
            try:
                method = next(method for method in clss.methods if method.sel.name == sel.name)
                break
            except StopIteration:
                pass

            clss = clss.superclass

        if method is None:
            # Could not resolve the method, either an invalid call, or we don't
            # know how to find out the method at runtime. Either way, not much
            # can be done.
            return None

        # Resolve name and number of function arguments from the selector.
        #
        # In Objective-C, the calling convention and number of arguments is
        # fixed at compile time, and no runtime reflection information is used,
        # so there's nothing stopping a program from using a selector that does
        # not at all reflect the real arguments to the method. Still, it is
        # conventional for the selector to be representative.
        #
        # Because of that, we try to extract some information only on a best-
        # effort basis, since we don't know whether it will be truly useful, or
        # if the program is actively trying to confuse us.
        sel_args: list[bytes] = [b"id", b"sel"]
        sel_last_args_idx = 0
        while len(sel_args) < max_method_argument_count.value:
            index = sel.name.find(b":", sel_last_args_idx)
            if index == -1:
                break

            sel_args.append(sel.name[sel_last_args_idx:index])
            sel_last_args_idx = index + 1

        # Resolve type and number of function arguments from the encoded type.
        #
        # Same caveats apply here as do with the selector.
        types = list(_parse_method_type_string(method.types))
        fn_rettype = types[0] if len(types) > 0 else None

        # Build the function using the information we got.
        fn_args: list[pwndbg.lib.functions.Argument] = []
        fn_args_unk_count = 0
        for arg_i in range(max(len(types) - 1, len(sel_args))):
            if arg_i < len(sel_args):
                name = sel_args[arg_i].decode("utf-8", errors="backslashreplace")
            else:
                # Name all arguments we don't have a name for "unknownX:"
                name = f"unknown{fn_args_unk_count}"
                fn_args_unk_count += 1

            if arg_i < len(types) - 1:
                ty = types[arg_i + 1].name_to_human_readable
            else:
                # Treat all arguments we don't know about as being `uintptr_t`s.
                ty = "uintptr_t"

            fn_args.append(pwndbg.lib.functions.Argument(type=ty, name=name, derefcnt=0))

        return pwndbg.lib.functions.Function(
            type=fn_rettype.name_to_human_readable if fn_rettype is not None else "void",
            derefcnt=0,
            name=sel.name.decode("utf-8", errors="backslashreplace"),
            args=fn_args,
        )

    # Not a an Objective-C call or not a type we know about.
    return None



from __future__ import annotations

import pwndbg.auxv
import pwndbg.chain
import pwndbg.commands
import pwndbg.lib.cache
from pwndbg.commands import CommandCategory


@pwndbg.commands.Command(
    "Print information from the Auxiliary ELF Vector.", category=CommandCategory.LINUX
)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.commands.OnlyWhenUserspace
def auxv() -> None:
    for k, v in pwndbg.auxv.get().items():
        if v is not None:
            print(k.ljust(24), v if not isinstance(v, int) else pwndbg.chain.format(v))


@pwndbg.commands.Command(
    "Explore and print information from the Auxiliary ELF Vector.", category=CommandCategory.LINUX
)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.commands.OnlyWhenUserspace
def auxv_explore() -> None:
    old_value = pwndbg.config.auto_explore_auxv.value
    pwndbg.config.auto_explore_auxv.value = "yes"
    try:
        pwndbg.lib.cache.clear_function_cache(pwndbg.auxv.get)
        pwndbg.auxv.get()
    finally:
        pwndbg.config.auto_explore_auxv.value = old_value

    auxv()



from __future__ import annotations

import typing
from collections import defaultdict
from enum import Enum
from typing import Protocol

import pwnlib
from capstone6pwndbg import *  # noqa: F403

# Reverse lookup tables for debug printing
from capstone6pwndbg import CS_AC
from capstone6pwndbg import CS_GRP
from capstone6pwndbg import CS_OP
from capstone6pwndbg.aarch64 import AARCH64_INS_BL
from capstone6pwndbg.aarch64 import AARCH64_INS_BLR
from capstone6pwndbg.aarch64 import AARCH64_INS_BR
from capstone6pwndbg.arm import ARM_INS_TBB
from capstone6pwndbg.arm import ARM_INS_TBH
from capstone6pwndbg.loongarch import LOONGARCH_INS_ALIAS_JR
from capstone6pwndbg.loongarch import LOONGARCH_INS_B
from capstone6pwndbg.loongarch import LOONGARCH_INS_BL
from capstone6pwndbg.loongarch import LOONGARCH_INS_CALL36
from capstone6pwndbg.loongarch import LOONGARCH_INS_JIRL
from capstone6pwndbg.mips import MIPS_INS_ALIAS_B
from capstone6pwndbg.mips import MIPS_INS_ALIAS_BAL
from capstone6pwndbg.mips import MIPS_INS_B
from capstone6pwndbg.mips import MIPS_INS_BAL
from capstone6pwndbg.mips import MIPS_INS_BLTZAL
from capstone6pwndbg.mips import MIPS_INS_J
from capstone6pwndbg.mips import MIPS_INS_JAL
from capstone6pwndbg.mips import MIPS_INS_JALR
from capstone6pwndbg.mips import MIPS_INS_JALR_HB
from capstone6pwndbg.mips import MIPS_INS_JR
from capstone6pwndbg.ppc import PPC_INS_B
from capstone6pwndbg.ppc import PPC_INS_BA
from capstone6pwndbg.ppc import PPC_INS_BL
from capstone6pwndbg.ppc import PPC_INS_BLA
from capstone6pwndbg.riscv import RISCV_INS_C_J
from capstone6pwndbg.riscv import RISCV_INS_C_JAL
from capstone6pwndbg.riscv import RISCV_INS_C_JALR
from capstone6pwndbg.riscv import RISCV_INS_C_JR
from capstone6pwndbg.riscv import RISCV_INS_JAL
from capstone6pwndbg.riscv import RISCV_INS_JALR
from capstone6pwndbg.sparc import SPARC_INS_ALIAS_BA
from capstone6pwndbg.sparc import SPARC_INS_ALIAS_CALL
from capstone6pwndbg.sparc import SPARC_INS_CALL
from capstone6pwndbg.sparc import SPARC_INS_JMPL
from capstone6pwndbg.systemz import SYSTEMZ_INS_B
from capstone6pwndbg.systemz import SYSTEMZ_INS_BAL
from capstone6pwndbg.systemz import SYSTEMZ_INS_BALR
from capstone6pwndbg.systemz import SYSTEMZ_INS_BR
from capstone6pwndbg.systemz import SYSTEMZ_INS_BRAS
from capstone6pwndbg.systemz import SYSTEMZ_INS_BRASL
from capstone6pwndbg.systemz import SYSTEMZ_INS_J
from capstone6pwndbg.systemz import SYSTEMZ_INS_JL
from capstone6pwndbg.x86 import X86_INS_CALL
from capstone6pwndbg.x86 import X86_INS_JMP
from capstone6pwndbg.x86 import X86_INS_RET
from capstone6pwndbg.x86 import X86Op
from typing_extensions import override

from pwndbg.dbg_mod import DisassembledInstruction

# Architecture specific instructions that mutate the instruction pointer unconditionally
UNCONDITIONAL_JUMP_INSTRUCTIONS: dict[int, set[int]] = {
    CS_ARCH_X86: {X86_INS_CALL, X86_INS_RET, X86_INS_JMP},
    CS_ARCH_MIPS: {
        MIPS_INS_J,
        MIPS_INS_JR,
        MIPS_INS_JAL,
        MIPS_INS_JALR,
        MIPS_INS_JALR_HB,
        MIPS_INS_BAL,
        MIPS_INS_ALIAS_BAL,
        MIPS_INS_B,
        MIPS_INS_ALIAS_B,
    },
    CS_ARCH_SPARC: {
        SPARC_INS_CALL,
        SPARC_INS_ALIAS_CALL,
        SPARC_INS_JMPL,
        SPARC_INS_ALIAS_BA,
    },
    CS_ARCH_ARM: {
        ARM_INS_TBB,
        ARM_INS_TBH,
    },
    CS_ARCH_AARCH64: {AARCH64_INS_BL, AARCH64_INS_BLR, AARCH64_INS_BR},
    CS_ARCH_RISCV: {
        RISCV_INS_JAL,
        RISCV_INS_JALR,
        RISCV_INS_C_JAL,
        RISCV_INS_C_JALR,
        RISCV_INS_C_J,
        RISCV_INS_C_JR,
    },
    CS_ARCH_PPC: {PPC_INS_B, PPC_INS_BA, PPC_INS_BL, PPC_INS_BLA},
    CS_ARCH_SYSTEMZ: {
        SYSTEMZ_INS_J,
        SYSTEMZ_INS_JL,
        SYSTEMZ_INS_B,
        SYSTEMZ_INS_BR,
        SYSTEMZ_INS_BAL,
        SYSTEMZ_INS_BALR,
        SYSTEMZ_INS_BRAS,
        SYSTEMZ_INS_BRASL,
    },
    CS_ARCH_LOONGARCH: {
        LOONGARCH_INS_B,
        LOONGARCH_INS_BL,
        LOONGARCH_INS_JIRL,
        LOONGARCH_INS_ALIAS_JR,
        LOONGARCH_INS_CALL36,
    },
}

# See: https://github.com/capstone-engine/capstone/issues/2448
BRANCH_AND_LINK_INSTRUCTIONS: dict[int, set[int]] = defaultdict(set)
BRANCH_AND_LINK_INSTRUCTIONS[CS_ARCH_MIPS] = {
    MIPS_INS_BAL,
    MIPS_INS_BLTZAL,
    MIPS_INS_JAL,
    MIPS_INS_JALR,
}

# All branch-like instructions - jumps thats are non-call and non-ret - should have one of these two groups in Capstone
GENERIC_JUMP_GROUPS = {CS_GRP_JUMP, CS_GRP_BRANCH_RELATIVE}

# All Capstone jumps should have at least one of these groups
ALL_JUMP_GROUPS = GENERIC_JUMP_GROUPS | {CS_GRP_CALL, CS_GRP_RET, CS_GRP_IRET}

# All non-ret jumps
FORWARD_JUMP_GROUP = {CS_GRP_CALL} | GENERIC_JUMP_GROUPS


class InstructionCondition(Enum):
    # Conditional instruction, and action is taken
    TRUE = 1
    # Conditional instruction, but action is not taken
    FALSE = 2
    # Unconditional instructions (most instructions), or we cannot reason about the instruction
    UNDETERMINED = 3


def boolean_to_instruction_condition(condition: bool) -> InstructionCondition:
    return InstructionCondition.TRUE if condition else InstructionCondition.FALSE


class SplitType(Enum):
    NO_SPLIT = 1
    BRANCH_TAKEN = 2
    BRANCH_NOT_TAKEN = 3


# Only use within the instruction.__repr__ to give a nice output
CAPSTONE_ARCH_MAPPING_STRING = {
    CS_ARCH_ARM: "arm",
    CS_ARCH_AARCH64: "aarch64",
    CS_ARCH_X86: "x86",
    CS_ARCH_PPC: "powerpc",
    CS_ARCH_MIPS: "mips",
    CS_ARCH_SPARC: "sparc",
    CS_ARCH_RISCV: "RISCV",
    CS_ARCH_SYSTEMZ: "s390x",
    CS_ARCH_LOONGARCH: "loongarch",
}


# Interface for enhanced instructions - there are two implementations defined in this file
class PwndbgInstruction(Protocol):
    cs_insn: CsInsn
    address: int
    size: int
    mnemonic: str
    op_str: str
    groups: set[int]
    id: int
    operands: list[EnhancedOperand]
    asm_string: str
    next: int
    target: int
    target_string: str | None
    target_const: bool | None
    condition: InstructionCondition
    declare_is_unconditional_jump: bool
    force_unconditional_jump_target: bool
    annotation: str | None
    annotation_padding: int | None
    syscall: int | None
    syscall_name: str | None
    causes_branch_delay: bool
    split: SplitType
    emulated: bool
    register_writes: dict[int, int]

    @property
    def call_like(self) -> bool: ...

    @property
    def jump_like(self) -> bool: ...

    @property
    def has_jump_target(self) -> bool: ...

    @property
    def is_conditional_jump(self) -> bool: ...

    @property
    def is_unconditional_jump(self) -> bool: ...

    @property
    def is_conditional_jump_taken(self) -> bool: ...

    @property
    def jump_result_is_known(self) -> bool: ...

    @property
    def bytes(self) -> bytearray: ...

    def op_find(self, op_type: int, position: int) -> EnhancedOperand: ...

    def op_count(self, op_type: int) -> int: ...


# This class is used to provide context to an instructions execution, used both
# in the disasm view output (see 'pwndbg.color.disasm.instruction()'), as well as for
# Pwndbg commands like "nextcall" that need to know the instructions target to set breakpoints
# The information in this class is backed by metadata from Capstone
class PwndbgInstructionImpl(PwndbgInstruction):
    def __init__(self, cs_insn: CsInsn, padding: int = 6) -> None:
        self.cs_insn: CsInsn = cs_insn
        """
        The underlying Capstone instruction object.
        Only the enhancement code should access the 'cs_insn' property
        """

        self.address: int = cs_insn.address

        self.size: int = cs_insn.size
        """
        Length of the instruction
        """

        self.mnemonic: str = cs_insn.mnemonic
        """
        Ex: 'MOV'
        """

        self.op_str: str = cs_insn.op_str
        """
        Ex: 'RAX, RDX'
        """

        self.groups: set[int] = set(cs_insn.groups)
        """
        Capstone instruction groups that we belong to.
        Groups that apply to all architectures: CS_GRP_INVALID | CS_GRP_JUMP | CS_GRP_CALL | CS_GRP_RET | CS_GRP_INT | CS_GRP_IRET | CS_GRP_PRIVILEGE | CS_GRP_BRANCH_RELATIVE
        """

        self.id: int = cs_insn.alias_id if cs_insn.is_alias else cs_insn.id
        """
        The underlying Capstone ID for the instruction
        If it's an alias, use the id of the alias

        Examples: X86_INS_JMP, X86_INS_CALL, RISCV_INS_C_JAL
        """

        # For ease, for x86 we will assume Intel syntax (destination operand first).
        # However, Capstone will disassemble using the `set disassembly-flavor` preference,
        # and the order of operands are read left to right into the .operands array. So we flip operand order if AT&T
        if self.cs_insn._cs.arch == CS_ARCH_X86 and self.cs_insn._cs.syntax == CS_OPT_SYNTAX_ATT:
            self.cs_insn.operands.reverse()

        self.operands: list[EnhancedOperand] = [EnhancedOperand(op) for op in self.cs_insn.operands]

        # ***********
        # The following member variables are set during instruction enhancement
        # in pwndbg.aglib.disasm.arch.py
        # ***********

        self.asm_string: str = (
            f"{self.mnemonic:<{padding}} {self.op_str}"
            if self.op_str
            else f"{self.mnemonic:<{padding}}"
        )
        """
        The full string representing the instruction - `mov    rdi, rsp` with appropriate padding.

        This is syntax highlighted during enhancement.

        This is additionally modified during enhancement for the purposes of replacing
        immediate values with their corresponding symbols
        """

        self.next: int = self.address + self.size
        """
        This is the address that the instruction pointer will be set to after using the "nexti" GDB command.
        This means it is the address of the next instruction to be executed in all cases except "call" instructions.

        Typically, it is `self.address + self.size` (the next instruction in memory)

        If it is a jump and we know it is taken, then it is the value of the jump target.

        Not set to "call" instruction targets, to indicate we will eventually (probably) return to this address
        """

        self.target: int = None
        """
        This is target of instructions that change the PC, regardless of if it's conditional or not,
        and whether or not we take the jump. This includes "call" and all other instructions that set the PC

        If the instruction is not one that changes the PC, target is set to "next"
        """

        self.target_string: str | None = None
        """
        String representation of the target address.

        Colorized symbol if a symbol exists at address, else colorized address
        """

        self.target_const: bool | None = None
        """
        Whether the target is a constant expression
        """

        self.condition: InstructionCondition = InstructionCondition.UNDETERMINED
        """
        Does the condition that the instruction checks for pass?

        For example, "JNE" jumps if Zero Flag is 0, else it does nothing. "CMOVA" conditionally performs a move depending on a flag.
        See 'condition' function in pwndbg.aglib.disasm.x86 for example on setting this.

        UNDETERMINED if we cannot reason about the condition, or if the instruction always executes unconditionally (most instructions).

        TRUE if the instruction has a conditional action, and we determine it is taken.

        FALSE if the instruction has a conditional action, and we know it is not taken.
        """

        self.declare_is_unconditional_jump: bool = False
        """
        This field is used to declare that this instruction is an unconditional jump.
        Most of the time, we depend on Capstone groups to check for jump instructions.
        However, some instructions become branches depending on the operands,
        such as Arm `add`, `sub`, `ldr`, `pop`, where PC is the destination register

        In these cases, we want to forcefully state that this instruction mutates the PC, so we set this attribute to True.

        This helps in two cases:
        1. Disassembly splits
        2. Instructions like `stepuntilasm` work better, as they detect these as branches to stop at.
        """

        self.force_unconditional_jump_target: bool = False
        """
        This asserts that the .target attribute is the real target of the instruction.
        This is only relevent in the edge case that the target is the next instruction in memory (address + size).
        The normal check for "target" checks that the target is NOT the next address in memory, and here we can assert that even if that is the case,
        we know that the jump really does just go to where self.target is.
        """

        self.annotation: str | None = None
        """
        The string is set in the "DisassemblyAssistant.enhance" function.
        It is used in the disasm print view to add context to the instruction, mostly operand value.
        This string is not used for all cases - if the instruction is a call or a jump, the 'target'.
        variables is used instead. See 'pwndbg.color.disasm.instruction()' for specific usage.
        """

        self.annotation_padding: int | None = None
        """
        The left adjustment padding that was used to previously print this.
        We retain it so the output is consistent between prints
        """

        self.syscall: int | None = None
        """
        The syscall number for this instruction, if it is a syscall. Otherwise None.
        """

        self.syscall_name: str | None = None
        """
        The syscall name as a string

        Ex: "openat", "read"
        """

        self.causes_branch_delay: bool = False
        """
        Whether or not this instruction has a single branch delay slot
        """

        self.split: SplitType = SplitType.NO_SPLIT
        """
        The type of split in the disasm display this instruction causes:

            NO_SPLIT            - no extra spacing between this and the next instruction
            BRANCH_TAKEN        - a newline with an arrow pointing down
            BRANCH_NOT_TAKEN    - an empty newline
        """

        self.emulated: bool = False
        """
        If the enhancement successfully used emulation for this instruction
        """

        self.register_writes = {}
        """
        Mapping of Capstone register id to integer value. During enhancement, we might manually determine
        that an instruction writes some value to a register, and this is stored here.
        """

    @property
    def call_like(self) -> bool:
        """
        True if this is a call-like instruction, meaning either it's a CALL or a branch and link.

        Checking for the CS_GRP_CALL is insufficient, as there are many "branch and link" instructions that are not labeled as a call
        """
        return (
            CS_GRP_CALL in self.groups
            or self.id in BRANCH_AND_LINK_INSTRUCTIONS[self.cs_insn._cs.arch]
        )

    @property
    def jump_like(self) -> bool:
        """
        True if this instruction is "jump-like", such as a JUMP, CALL, or RET.
        Basically, the PC is set to some target by means of this instruction.

        It may still be a conditional jump - this property does not indicate whether the jump is taken or not.
        """
        return bool(self.groups & ALL_JUMP_GROUPS) or self.declare_is_unconditional_jump

    @property
    def has_jump_target(self) -> bool:
        """
        True if we have determined that this instruction can explicitly change the program counter, and
        we have determined the jump target.

        Edge case - the jump target MAY be the next address in memory - so we check force_unconditional_jump_target
        """
        # The second check ensures that if the target address is itself, it's a jump (infinite loop) and not something like `rep movsb` which repeats the same instruction.
        # Because capstone doesn't catch ALL cases of an instruction changing the PC, we don't have the `jump_like` in the first part of this check.
        return (
            self.target not in (None, self.address + self.size)
            and (self.target != self.address or self.jump_like)
        ) or self.force_unconditional_jump_target

    @property
    def is_conditional_jump(self) -> bool:
        """
        True if this instruction can change the program counter conditionally.

        This is used, in part, to determine if the instruction deserves a "checkmark" in the disasm view.

        This does not imply that we have resolved the .target
        """
        return (
            self.declare_is_unconditional_jump is False
            and bool(self.groups & GENERIC_JUMP_GROUPS)
            and self.id not in UNCONDITIONAL_JUMP_INSTRUCTIONS[self.cs_insn._cs.arch]
        )

    @property
    def is_unconditional_jump(self) -> bool:
        """
        True if we know the instruction can change the program counter, and does so unconditionally.

        This includes things like RET, CALL, and JMP (in x86).

        This property is used in enhancement to determine certain codepaths when resolving .next for this instruction.

        This does not imply that we have resolved the .target
        """
        return (
            self.declare_is_unconditional_jump
            or self.id in UNCONDITIONAL_JUMP_INSTRUCTIONS[self.cs_insn._cs.arch]
        )

    @property
    def is_conditional_jump_taken(self) -> bool:
        """
        True if this is a conditional jump, and we predicted that we will take the jump
        """
        # True if:
        # - We manually determined in .condition that we take the jump
        # - Or that emulation determined the .next to go somewhere and we didn't explicitely set .condition to False.
        #   Emulation can be incorrect, so we check the conditional for false to check if we manually override the emulator's decision
        return self.is_conditional_jump and (
            self.condition == InstructionCondition.TRUE
            or (
                (self.next not in (None, self.address + self.size))
                and self.condition != InstructionCondition.FALSE
            )
        )

    @property
    def jump_result_is_known(self) -> bool:
        """
        True under the following conditions:
        - If it's an unconditional jump, we know the target of the jump
        - If it's a conditional jump, we know the target of the branch and know whether or not we take it
        Otherwise, false
        """
        return self.has_jump_target and (
            (self.is_unconditional_jump)
            or (self.is_conditional_jump and self.condition != InstructionCondition.UNDETERMINED)
        )

    @property
    def bytes(self) -> bytearray:
        """
        Raw machine instruction bytes
        """
        return self.cs_insn.bytes

    def op_find(self, op_type: int, position: int) -> EnhancedOperand:
        """Get the operand at position @position of all operands having the same type @op_type"""
        cs_op = self.cs_insn.op_find(op_type, position)
        # Find the matching EnhancedOperand
        for x in self.operands:
            if x.cs_op == cs_op:
                return x
        return None

    def op_count(self, op_type: int) -> int:
        """Return number of operands having same operand Capstone type 'op_type'"""
        return self.cs_insn.op_count(op_type)

    # For debugging
    def __repr__(self) -> str:
        operands_str = " ".join([repr(op) for op in self.operands])

        hex_register_writes = {
            self.cs_insn.reg_name(reg_id): hex(reg_value)
            for reg_id, reg_value in self.register_writes.items()
        }

        info = f"""{self.mnemonic} {self.op_str} at {self.address:#x} (size={self.size}) (arch: {CAPSTONE_ARCH_MAPPING_STRING.get(self.cs_insn._cs.arch, None)})
        Bytes: {pwnlib.util.fiddling.enhex(self.bytes)}
        ID: {self.id}, {self.cs_insn.insn_name()}
        Capstone ID/Alias ID: {self.cs_insn.id} / {self.cs_insn.alias_id if self.cs_insn.is_alias else "None"}
        Raw asm: f"{self.mnemonic:-<6} {self.op_str}"
        New asm: {self.asm_string}
        Next: {self.next:#x}
        Target: {hex(self.target) if self.target is not None else None}, Target string={self.target_string or ""}, const={self.target_const}
        Condition: {self.condition.name}
        Groups: {[CS_GRP.get(group, group) for group in self.groups]}
        Annotation: {self.annotation}
        Operands: [{operands_str}]
        Conditional jump: {self.is_conditional_jump}. Taken: {self.is_conditional_jump_taken}
        Unconditional jump: {self.is_unconditional_jump}
        Declare unconditional jump: {self.declare_is_unconditional_jump}
        Force jump target: {self.force_unconditional_jump_target}
        Can change PC: {self.has_jump_target}
        Syscall: {self.syscall if self.syscall is not None else ""} {self.syscall_name if self.syscall_name is not None else "N/A"}
        Causes Delay slot: {self.causes_branch_delay}
        Split: {SplitType(self.split).name}
        Call-like: {self.call_like}
        Register writes: {hex_register_writes}"""

        try:
            regs_read, regs_written = self.cs_insn.regs_access()
            info += f"\n\tCapstone regs read: {[self.cs_insn.reg_name(reg) for reg in regs_read]}"
            info += (
                f"\n\tCapstone regs written: {[self.cs_insn.reg_name(reg) for reg in regs_written]}"
            )
        except CsError:
            # Not all architectures support the .reg_access() API
            pass

        # Hacky, but this is just for debugging
        if hasattr(self.cs_insn, "cc"):
            info += f"\n\tARM condition code: {self.cs_insn.cc}"
            info += f"\n\tThumb mode: {1 if self.cs_insn._cs._mode & CS_MODE_THUMB else 0}"

        if hasattr(self.cs_insn, "cc_field"):
            info += f"\n\tSPARC cc_field: {self.cs_insn.cc_field}"

        return info


class EnhancedOperand:
    def __init__(self, cs_op):
        self.cs_op: typing.Any = cs_op
        """
        Underlying Capstone operand. Takes on a different value depending on the architecture.

        x86 = capstone.x86.X86Op, arm = capstone.arm.ArmOp, mips = capstone.mips.MipsOp
        """

        # ***********
        # The following member variables are set during instruction enhancement
        # in pwndbg.aglib.disasm.arch.py
        # ***********

        self.before_value: int | None = None
        """
        The value of the operand before the instruction executes.
        This is set only if the operand value can be reasoned about.
        """

        self.after_value: int | None = None
        """
        The value of the operand after the instruction executes.
        Only set when using emulation.
        """

        self.before_value_resolved: int | None = None
        """
        The 'resolved' value of the operand that is actually used in the instruction logic, before the instruction executes.
        This is the same as before_value if it's not a memory operand, in which cases it's the dereferenced value.

        Helpful for cases like  `cmp    byte ptr [rip + 0x166669], 0`, where first operand could be
        a register or a memory value to dereference, and we want the actual value used.
        """

        self.before_value_no_modifiers: int | None = None
        """
        This is a special field used in some architectures that allow operand modifiers, such as shifts and extends in Arm.
        Capstone bundles the modifier with the operand, and when we are resolving concrete operand values, we apply the modifier.
        However, in some annotations we need to un-modified raw register value, which is what this field is for.
        """

        self.after_value_resolved: int | None = None
        """
        The 'resolved' value of the operand after the instruction executes.
        """

        self.str: str | None = ""
        """
        String representing the operand

        Ex: "RAX", or "[0x7fffffffd9e8]". None if value cannot be determined.
        """

        self.symbol: str | None = None
        """
        Colorized symbol name for this operand, if .before_value is set and symbol exists, else None.
        """

    @property
    def type(self) -> int:
        """
        CS_OP_REG | CS_OP_MEM | CS_OP_IMM | CS_OP_INVALID | CS_OP_FP
        """
        return self.cs_op.type

    @property
    def reg(self) -> int:
        """
        The underlying Capstone ID for the register
        """
        return self.cs_op.reg

    @property
    def imm(self) -> int:
        """
        The immediate value of the operand (if applicable)
        """
        return self.cs_op.imm

    @property
    def mem(self) -> typing.Any:
        """
        Return the underlying Capstone mem object (if applicable)
        """
        return self.cs_op.value.mem

    # For debugging
    def __repr__(self) -> str:
        info = (
            f"'{self.str}': Symbol: {self.symbol}, "
            f"Before: {hex(self.before_value) if self.before_value is not None else None}, "
            f"After: {hex(self.after_value) if self.after_value is not None else None}, "
            f"type={CS_OP.get(self.type, self.type)}"
        )

        if isinstance(self.cs_op, X86Op):
            info += (
                f", size={self.cs_op.size}, "
                f"access={CS_AC.get(self.cs_op.access, self.cs_op.access)}]"
            )

        return f"[{info}]"


# Represents a disassembled instruction
# Conforms to the PwndbgInstruction interface
class ManualPwndbgInstruction(PwndbgInstruction):
    def __init__(self, address: int, padding: int = 6) -> None:
        """
        This class provides an implementation of PwndbgInstruction for cases where the architecture
        at hand is not supported by the Capstone disassembler. The backing information is sourced from
        GDB/LLDB's built-in disassemblers.

        Instances of this class do not go through the 'enhancement' process due to lacking important information provided by Capstone.
        As a result of this, some of the methods raise NotImplementedError, because if they are called it indicates a bug elsewhere in the codebase.
        """
        ins: DisassembledInstruction = pwndbg.dbg.selected_inferior().disasm(address)
        asm = ins["asm"].split(maxsplit=1)

        # The enhancement code assumes this value exists.
        # However, a ManualPwndbgInstruction should never be used in the enhancement code.
        self.cs_insn: CsInsn = None

        self.address = address
        self.size = ins["length"]

        self.mnemonic = asm[0].strip()
        self.op_str = asm[1].strip() if len(asm) > 1 else ""
        self.groups = set()

        # Set Capstone ID to -1
        self.id = -1

        self.operands = []

        self.asm_string = (
            f"{self.mnemonic:<{padding}} {self.op_str}"
            if self.op_str
            else f"{self.mnemonic:<{padding}}"
        )

        self.next = address + self.size
        self.target = self.next
        self.target_string = None
        self.target_const = None

        self.condition = InstructionCondition.UNDETERMINED

        self.declare_is_unconditional_jump = False
        self.force_unconditional_jump_target = False

        self.annotation = None

        self.annotation_padding = None

        self.syscall = None
        self.syscall_name = None

        self.causes_branch_delay = False

        self.split = SplitType.NO_SPLIT

        self.emulated = False

        self.register_writes = {}

    @property
    def bytes(self) -> bytearray:
        # GDB simply doesn't provide us with the raw bytes.
        # However, it is important that this returns a valid bytearray,
        # since the disasm code indexes this for nearpc-num-opcode-bytes.
        return bytearray()

    @property
    def call_like(self) -> bool:
        return False

    @property
    def jump_like(self) -> bool:
        return False

    @property
    def has_jump_target(self) -> bool:
        return False

    @property
    def is_conditional_jump(self) -> bool:
        return False

    @property
    def is_unconditional_jump(self) -> bool:
        return False

    @property
    def is_conditional_jump_taken(self) -> bool:
        return False

    @property
    def jump_result_is_known(self) -> bool:
        return False

    @override
    def op_find(self, op_type: int, position: int) -> EnhancedOperand:
        # raise NotImplementedError, because if this is called it indicates a bug elsewhere in the codebase.
        # ManualPwndbgInstruction should not go through the enhancement process, where this would be called.
        raise NotImplementedError

    @override
    def op_count(self, op_type: int) -> int:
        raise NotImplementedError



from __future__ import annotations

from collections.abc import Callable
from collections.abc import Sequence
from typing import Any

import pwndbg.color
from pwndbg import config
from pwndbg.lib.config import Parameter
from pwndbg.lib.config import Scope


class ColorParameter(Parameter):
    color_function: Callable[[object], str]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.update_color_function()

    def update_color_function(self):
        self.color_function = pwndbg.color.generateColorFunction(self.value)


def add_param(
    name: str,
    default: Any,
    set_show_doc: str,
    *,
    help_docstring: str = "",
    param_class: int | None = None,
    enum_sequence: Sequence[str] | None = None,
) -> Parameter:
    return config.add_param(
        name,
        default,
        set_show_doc,
        scope=Scope.theme,
        help_docstring=help_docstring,
        param_class=param_class,
        enum_sequence=enum_sequence,
    )


def add_color_param(
    name: str, default: Any, set_show_doc: str, *, help_docstring: str = ""
) -> ColorParameter:
    color_parameter = ColorParameter(
        name, default, set_show_doc, help_docstring=help_docstring, scope=Scope.theme
    )

    config.triggers[name].append(color_parameter.update_color_function)

    config.add_param_obj(color_parameter)

    return color_parameter



"""
Debugger-agnostic library that provides various functionality.

Takes the debugging primitives provided by the Debugger API
and builds the more complex and interesting bits of functionality
found in Pwndbg on top of them.

See https://pwndbg.re/dev/contributing/dev-notes/#aglib
for more information.
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

# Since pwndbg.aglib.whatever is a common pattern in the aglib/ files
# we have to make sure that python can quickly initialize the pwndbg.aglib
# submodule without importing anything else (specifically, without importing
# any aglib file, or anything that depends on an aglib file, because we would
# get a circular import).

if TYPE_CHECKING:
    # Without ^this if-statement, we get a circular import.
    from pwndbg.aglib.arch_mod import PwndbgArchitecture
    from pwndbg.aglib.regs_mod import RegisterManager

    # These will be set during debugger setup.
    # Thus you can't import them with `from pwndbg.aglib import arch`.
    arch: PwndbgArchitecture = cast(PwndbgArchitecture, None)
    regs: RegisterManager = cast(RegisterManager, None)
else:
    arch = None
    regs = None


def load_aglib():
    # We need this for the reason commented above. If we changed the
    # pwndbg/aglib/ files to access their siblings via
    # `import .memory` instead of `import pwndbg.aglib.memory`
    # we could get rid of this function.

    import pwndbg.aglib.argv
    import pwndbg.aglib.dynamic
    import pwndbg.aglib.elf
    import pwndbg.aglib.file
    import pwndbg.aglib.heap
    import pwndbg.aglib.kernel
    import pwndbg.aglib.kernel.vmmap
    import pwndbg.aglib.macho
    import pwndbg.aglib.memory
    import pwndbg.aglib.nearpc
    import pwndbg.aglib.next
    import pwndbg.aglib.objc
    import pwndbg.aglib.onegadget
    import pwndbg.aglib.proc
    import pwndbg.aglib.qemu
    import pwndbg.aglib.regs_mod
    import pwndbg.aglib.remote
    import pwndbg.aglib.stack
    import pwndbg.aglib.strings
    import pwndbg.aglib.symbol
    import pwndbg.aglib.typeinfo
    import pwndbg.aglib.vmmap
    import pwndbg.aglib.vmmap_custom

    global regs
    regs = pwndbg.aglib.regs_mod.regs


def set_arch(pwndbg_arch: PwndbgArchitecture):
    global arch
    arch = pwndbg_arch



"""
Shellcode

This module implements functionality that allows for the execution of a small
amount of code in the context of the inferior.

"""

from __future__ import annotations

import gdb
import pwnlib.shellcraft

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.asm
import pwndbg.aglib.memory
import pwndbg.aglib.vmmap
import pwndbg.gdblib.prompt
import pwndbg.lib.regs


def _get_syscall_return_value():
    """
    Reads the value corresponding to the return value of a syscall that has
    just returned.
    """

    register_set = pwndbg.lib.regs.reg_sets[pwndbg.aglib.arch.name]
    return pwndbg.aglib.regs.read_reg(register_set.retval)


def exec_syscall(
    syscall,
    arg0=None,
    arg1=None,
    arg2=None,
    arg3=None,
    arg4=None,
    arg5=None,
    arg6=None,
    disable_breakpoints=False,
):
    """
    Tries executing the given syscall in the context of the inferior.
    """

    # Build machine code that runs the requested syscall.
    syscall_asm = pwnlib.shellcraft.syscall(syscall, arg0, arg1, arg2, arg3, arg4, arg5)
    syscall_bin = pwndbg.aglib.asm.asm(syscall_asm)

    # Run the syscall and pass its return value onward to the caller.
    return exec_shellcode(
        syscall_bin,
        restore_context=True,
        capture=_get_syscall_return_value,
        disable_breakpoints=disable_breakpoints,
    )


def exec_shellcode(blob, restore_context=True, capture=None, disable_breakpoints=False):
    """
    Tries executing the given blob of machine code in the current context of the
    inferior, optionally restoring the values of the registers as they were
    before the shellcode ran, as a means to allow for execution of the inferior
    to continue uninterrupted. The value of the program counter is always
    restored.

    Additionally, the caller may specify an object to be called before the
    context is restored, so that information stored in the registers after the
    shellcode finishes can be retrieved. The return value of that call will be
    returned by this function.

    # Safety
    Seeing as this function injects code directly into the inferior and runs it,
    the caller must be careful to inject code that will (1) terminate and (2)
    not cause the inferior to misbehave. Otherwise, it is fairly easy to crash
    or currupt the memory in the inferior.
    """

    register_set = pwndbg.lib.regs.reg_sets[pwndbg.aglib.arch.name]
    preserve_set = register_set.gpr + register_set.args + (register_set.pc, register_set.stack)

    registers = {reg: pwndbg.aglib.regs.read_reg(reg) for reg in preserve_set}
    starting_address = registers[register_set.pc]

    # Make sure the blob fits in the rest of the space we have in this page.
    #
    # NOTE: Technically, we could actually use anything from the whole page to
    # all of the pages currently mapped as executable for this. There is no
    # technical limitation stopping us from doing that, but seeing as doing it
    # is harder to make sure it works correctly, we don't (for now, at least).
    page = pwndbg.aglib.vmmap.find(starting_address)
    assert page is not None

    clearance = page.end - len(blob) - 1
    if clearance < 0:
        # The page isn't large enough to hold our shellcode.
        raise RuntimeError(
            f"Not enough space to execute code as inferior: \
            need at least {len(blob)} bytes, have {clearance} bytes available"
        )

    # Swap the code in the range with our shellcode.
    existing_code = pwndbg.aglib.memory.read(starting_address, len(blob))
    pwndbg.aglib.memory.write(starting_address, blob)

    # Disable breakpoints.
    #
    # We might not want to hit any breakpoints that aren't the ones that are
    # strictly necessary for this function to run to completion. So, if
    # requested, we temporarily disable breakpoints during the shellcode
    re_enable_list = []
    for bp in gdb.breakpoints():
        if bp.enabled and disable_breakpoints:
            bp.enabled = False
            re_enable_list.append(bp)

    # The continue we use here will trigger an event that would get the context
    # prompt to show, regardless of the circumstances. We don't want that, so
    # we preserve the state of the context skip.
    would_skip_context = pwndbg.gdblib.prompt.context_shown

    # Execute.
    target_address = starting_address + len(blob)
    bp = gdb.Breakpoint(f"*{target_address:#x}", internal=True, temporary=True)
    bp.silent = True
    gdb.execute("continue")

    # Restore the state of the context skip.
    pwndbg.gdblib.prompt.context_shown = would_skip_context

    # Re-enable breakpoints.
    for bp in re_enable_list:
        bp.enabled = True

    # Make sure we're in the right place.
    assert pwndbg.aglib.regs.pc == target_address

    # Give the caller a chance to collect information from the environment
    # before any of the context gets restored.
    captured = None
    if capture is not None:
        captured = capture()

    # Restore the code and the program counter and, if requested, the rest of
    # the registers.
    pwndbg.aglib.memory.write(starting_address, existing_code)
    pwndbg.aglib.regs.write_reg(register_set.pc, starting_address)
    if restore_context:
        for reg, val in registers.items():
            pwndbg.aglib.regs.write_reg(reg, val)

    return captured



from __future__ import annotations

import argparse

import pwndbg.aglib
import pwndbg.color.memory as mem_color
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="Print the distance between the two arguments, or print the offset to the address's page base."
)
parser.add_argument("a", type=pwndbg.commands.AddressExpr, help="The first address.")
parser.add_argument(
    "b", nargs="?", default=None, type=pwndbg.commands.AddressExpr, help="The second address."
)


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
def distance(a, b) -> None:
    """Print the distance between the two arguments"""

    if b is None:
        page = pwndbg.aglib.vmmap.find(a)

        if not page:
            print(f"{a:#x} does not belong to a mapped page in memory")
        else:
            # a is a gdb.Value, explicitely convert to int
            distance = int(a) - page.vaddr

            display_text = f"{page.vaddr:#x}->{a:#x} is {distance:#x} bytes ({distance // pwndbg.aglib.arch.ptrsize:#x} words)"

            print(mem_color.get(page.vaddr, text=display_text))
    else:
        a = int(a) & pwndbg.aglib.arch.ptrmask
        b = int(b) & pwndbg.aglib.arch.ptrmask

        distance = b - a

        print(
            f"{a:#x}->{b:#x} is {distance:#x} bytes ({distance // pwndbg.aglib.arch.ptrsize:#x} words)"
        )



from __future__ import annotations

import argparse

import pwndbg.aglib.kernel
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(description="Outputs the kernel version (/proc/version).")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWhenPagingEnabled
def kversion() -> None:
    print(pwndbg.aglib.kernel.kversion())



from __future__ import annotations

import argparse
from collections.abc import Iterator

import pwndbg.aglib
import pwndbg.aglib.file
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.aglib.vmmap
import pwndbg.chain
import pwndbg.color.memory as mem_color
import pwndbg.commands
import pwndbg.lib.memory
import pwndbg.wrappers.checksec
import pwndbg.wrappers.readelf
from pwndbg.color import message
from pwndbg.commands import CommandCategory
from pwndbg.wrappers.readelf import RelocationType

parser = argparse.ArgumentParser(
    description="Show the state of the Global Offset Table.",
)
group = parser.add_mutually_exclusive_group()
group.add_argument(
    "-p",
    "--path",
    help="Filter results by library/objfile path.",
    type=str,
    default="",
    dest="path_filter",
)
group.add_argument(
    "-a",
    "--all",
    help="Process all libs/obfjiles including the target executable.",
    action="store_true",
    default=False,
    dest="all_",
)
parser.add_argument(
    "-r",
    "--show-readonly",
    help="Also display read-only entries (which are filtered out by default).",
    action="store_true",
    default=False,
    dest="accept_readonly",
)
parser.add_argument(
    "symbol_filter", help="Filter results by symbol name.", type=str, nargs="?", default=""
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.LINUX,
    examples="""
> got
    Print all writable GOT entries in the executable.
> got -r puts
    Print all GOT entries that contain the string "puts".
> got -p libc
    Print all writable GOT entries used by libc. (And any other loaded
    object files that contain the string "libc" in their path).
> got -ra
    Print all GOT entries in the address space.
""",
)
@pwndbg.commands.OnlyWhenRunning
def got(path_filter: str, all_: bool, accept_readonly: bool, symbol_filter: str) -> None:
    # Show the filters we are using
    if path_filter:
        print("Filtering by lib/objfile path: " + message.hint(path_filter))
    if symbol_filter:
        print("Filtering by symbol name: " + message.hint(symbol_filter))
    if not accept_readonly:
        print("Filtering out read-only entries (display them with -r or --show-readonly)")

    if path_filter or not accept_readonly or symbol_filter:
        print()

    # Calculate the base address
    if not path_filter:
        first_print = False
        _got(pwndbg.aglib.proc.exe(), accept_readonly, symbol_filter)
    else:
        first_print = True

    if not all_ and not path_filter:
        return

    paths = [o.objfile for o in iter_objfiles()]
    paths.sort()
    for path in paths:
        if path_filter not in path:
            continue
        if not first_print:
            print()
        first_print = False
        _got(path, accept_readonly, symbol_filter)

    # Maybe user have a typo or something in the path filter, show the available shared libraries
    if first_print and path_filter:
        print(message.error("No shared library matching the path filter found."))
        if paths:
            print(message.notice("Available shared libraries:"))
            for path in paths:
                print("    " + path)


def _got(path: str, accept_readonly: bool, symbol_filter: str) -> None:
    # Maybe download the file from remote
    local_path = pwndbg.aglib.file.get_file(path, try_local_path=True)

    relro_status = pwndbg.wrappers.checksec.relro_status(local_path)
    pie_status = pwndbg.wrappers.checksec.pie_status(local_path)
    got_entry = pwndbg.wrappers.readelf.get_got_entry(local_path)

    # The following code is inspired by the "got" command of https://github.com/bata24/gef/blob/dev/gef.py by @bata24, thank you!
    # TODO/FIXME: Maybe a -v option to show more information will be better
    outputs: list[dict[str, str | int]] = []
    if path == pwndbg.aglib.proc.exe():
        bin_base_offset = pwndbg.aglib.proc.binary_base_addr() if "PIE enabled" in pie_status else 0
    else:
        page = next(filter(lambda o: o.objfile == path, iter_objfiles()), None)
        assert page is not None, f"unable to find vmmap entry for objfile: {path}"
        bin_base_offset = page.start

    # Parse the output of readelf line by line
    for category, lines in got_entry.items():
        for line in lines:
            # There are 5 fields in the output of readelf:
            # "Offset", "Info", "Type", "Sym. Value", and "Symbol's Name"
            # We only care about "Offset", "Sym. Value" and "Symbol's Name" here
            offset, _, _, *rest = line.split()[:5]
            if len(rest) < 2:
                # "Sym. Value" or "Symbol's Name" are not present in this case
                # The output of readelf might look like this (missing both value and name):
                # 00004e88  00000008 R_386_RELATIVE
                # or something like this (only missing name):
                # 00000000001ec018  0000000000000025 R_X86_64_IRELATIVE                        a0480
                # TODO: Is it possible that we are missing the value but not the name?
                value = rest[0] if rest else ""
                name = ""
            else:
                # Every fields are present in this case
                # The output of readelf might look like this:
                # 00000000001ec030  0000020a00000007 R_X86_64_JUMP_SLOT     000000000009ae80 realloc@@GLIBC_2.2.5 + 0
                value, name = rest
            address = int(offset, 16) + bin_base_offset
            # TODO/FIXME: This check might not work correctly if we failed to get the correct vmmap result
            if not accept_readonly and not pwndbg.aglib.vmmap.find(address).write:
                continue
            if not name and category == RelocationType.IRELATIVE:
                # TODO/FIXME: I don't know the naming logic behind this yet, I'm just modifying @bata24's code here :p
                # We might need to add some comments here to explain the logic in the future, and also fix it if something wrong
                if pwndbg.aglib.arch.name == "i386":
                    name = "*ABS*"
                else:
                    name = f"*ABS*+0x{int(value, 16):x}"
            if symbol_filter not in name:
                continue
            outputs.append(
                {
                    "name": name or "????",
                    "address": address,
                }
            )
    # By sorting the outputs by address, we can get a more intuitive output
    outputs.sort(key=lambda x: x["address"])
    relro_color = message.off
    if "Partial" in relro_status:
        relro_color = message.warn
    elif "Full" in relro_status:
        relro_color = message.on
    print(f"State of the GOT of {message.notice(path)}:")
    print(
        f"GOT protection: {relro_color(relro_status)} | Found {message.hint(len(outputs))} GOT entries passing the filter"
    )
    for output in outputs:
        print(
            f"[{mem_color.get(output['address'])}] {message.hint(output['name'])} -> {pwndbg.chain.format(pwndbg.aglib.memory.read_pointer_width(output['address']))}"  # type: ignore[arg-type]
        )


def iter_objfiles() -> Iterator[pwndbg.lib.memory.Page]:
    main = pwndbg.aglib.proc.exe()
    uniq = set()

    for page in pwndbg.aglib.vmmap.get():
        if page.objfile == main:
            # Skip main elf
            continue
        if not page.is_memory_mapped_file:
            # Skip virtual objfiles eg: `[vdso]` etc..
            continue
        if page.objfile in uniq:
            continue
        uniq.add(page.objfile)
        yield page



from __future__ import annotations

import argparse
import math
import os

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.vmmap
import pwndbg.color.memory as mem_color
import pwndbg.commands
import pwndbg.dbg_mod
from pwndbg.color import message
from pwndbg.commands import CommandCategory


def find_module(addr, max_distance):
    mod_filter = lambda page: page.start <= addr < page.end
    pages = list(filter(mod_filter, pwndbg.aglib.vmmap.get()))

    if not pages:
        if max_distance != 0:
            mod_filter = lambda page: page.start - max_distance <= addr < page.end + max_distance
            pages = list(filter(mod_filter, pwndbg.aglib.vmmap.get()))

        if not pages:
            return None

    return pages[-1]


def satisfied_flags(require_flags, flags):
    return (require_flags & ~(flags)) == 0


def flags_str2int(flags_s):
    flag_i = 0
    if "r" in flags_s:
        flag_i |= os.R_OK
    if "w" in flags_s:
        flag_i |= os.W_OK
    if "x" in flags_s:
        flag_i |= os.X_OK
    return flag_i


parser = argparse.ArgumentParser(
    description="Pointer scan for possible offset leaks.",
)
parser.add_argument(
    "address",
    nargs="?",
    type=int,
    # Legal because it will get parsed by the debugger.
    default="$sp",
    help="Leak memory address",
)
parser.add_argument("count", nargs="?", default=0x40, type=int, help="Leak size in bytes")
parser.add_argument(
    "--max-distance",
    type=int,
    default=0x0,
    help="Max acceptable distance between memory page boundary and leaked pointer",
)
parser.add_argument(
    "--point-to",
    type=str,
    default=None,
    help="Mapping name of the page that you want the pointers point to",
)
parser.add_argument(
    "--max-ptrs", type=int, default=0, help="Stop search after find n pointers, default 0"
)
parser.add_argument(
    "--flags",
    type=str,
    default=None,
    help="flags of the page that you want the pointers point to. [e.g. rwx]",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MEMORY,
    examples="""
> probeleak $rsp 0x64
Leaks 0x64 bytes starting at stack pointer and search for valid pointers.
> probeleak $rsp 0x64 --max-dist 0x10
As above, but pointers may point 0x10 bytes outside of memory page.
> probeleak $rsp 0x64 --point-to libc --max-ptrs 1 --flags rwx
Leaks 0x64 bytes starting at stack pointer and search for one valid pointer
which points to a libc rwx page.
""",
)
@pwndbg.commands.OnlyWhenRunning
def probeleak(
    address=None, count=0x40, max_distance=0x0, point_to=None, max_ptrs=0, flags=None
) -> None:
    address = int(address)
    address &= pwndbg.aglib.arch.ptrmask
    ptrsize = pwndbg.aglib.arch.ptrsize
    count = max(int(count), ptrsize)
    off_zeros = int(math.ceil(math.log(count, 2) / 4))
    if flags is not None:
        require_flags = flags_str2int(flags)

    if count > address > 0x10000:  # in case someone puts in an end address and not a count (smh)
        print(
            message.warn(
                f"Warning: you gave an end address, not a count. Subtracting 0x{address:x} from the count."
            )
        )
        count -= address

    try:
        data = pwndbg.aglib.memory.read(address, count, partial=True)
    except pwndbg.dbg_mod.Error as e:
        print(message.error(str(e)))
        return

    if not data:
        print(
            message.error(
                f"Couldn't read memory at 0x{address:x}. See 'probeleak -h' for the usage."
            )
        )
        return

    found = False
    find_cnt = 0
    for i in range(0, len(data) - ptrsize + 1):
        p = pwndbg.aglib.arch.unpack(data[i : i + ptrsize])
        page = find_module(p, max_distance)
        if page:
            if point_to is not None and point_to not in page.objfile:
                continue
            if flags is not None and not satisfied_flags(require_flags, page.flags):
                continue
            if not found:
                print(mem_color.legend())
                found = True

            mod_name = page.objfile
            if not mod_name:
                mod_name = "[anon]"

            if p >= page.end:
                right_text = f"({page.permstr}) {mod_name} + 0x{page.memsz:x} + 0x{p - page.end:x} (outside of the page)"
            elif p < page.start:
                right_text = (
                    f"({page.permstr}) {mod_name} - 0x{page.start - p:x} (outside of the page)"
                )
            else:
                right_text = f"({page.permstr}) {mod_name} + 0x{p - page.start:x}"

            offset_text = f"0x{i:0{off_zeros}x}"
            p_text = f"0x{p:0{int(ptrsize * 2)}x}"
            text = f"{offset_text}: {mem_color.get(p, text=p_text)} = {mem_color.get(p, text=right_text)}"

            symbol = pwndbg.aglib.symbol.resolve_addr(p)
            if symbol:
                text += f" ({symbol})"
            print(text)

            find_cnt += 1
            if max_ptrs != 0 and find_cnt >= max_ptrs:
                break

    if not found:
        print(message.hint(f"No leaks found at 0x{address:x}-0x{address + count:x} :("))



from __future__ import annotations

from re import match
from re import search
from struct import unpack_from

from pwnlib.util.packing import p16
from pwnlib.util.packing import u32
from pwnlib.util.packing import u64

import pwndbg.aglib
import pwndbg.aglib.kernel
import pwndbg.aglib.kernel.kmod
import pwndbg.aglib.memory
import pwndbg.color.message as message
import pwndbg.lib.cache


@pwndbg.lib.cache.cache_until("start")
def get() -> tuple[tuple[str, str, int], ...]:
    ks = Kallsyms()
    return tuple(ks.kallsyms)


class Kallsyms:
    """
    - linux_banner >= 6.4
    - ... <= 6.4
    - kallsyms_offsets
    - kallsyms_relative_base
    - kallsyms_num_syms
    - kallsyms_names
    - kallsyms_markers
    - kallsyms_token_table
    - kallsyms_token_index
    - kallsyms_offsets >= 6.4
    - kallsyms_relative_base >= 6.4
    """

    def __init__(self):
        self.kallsyms: list[tuple[str, str, int]] = []
        self.kbase = pwndbg.aglib.kernel.kbase()
        if self.kbase is None:
            print(message.warn("could not find kbase, kernel has not finished initialization?"))
            return

        mapping = pwndbg.aglib.kernel.first_kernel_ro_page()
        assert mapping is not None, "kernel memory mappings are missing"

        self.r_base = mapping.vaddr
        self.kernel_ro_mem = pwndbg.aglib.memory.read(mapping.vaddr, mapping.memsz)

        self.kernel_version = pwndbg.aglib.kernel.krelease()
        self.is_offsets = False

        self.rbase_offset = 0

        self.is_big_endian = None

        self.token_table = self.find_token_table()
        self.is_uncompressed = False

        if self.token_table is None:
            if not self.find_names_uncompressed():
                return
            print(message.info("Detected Uncompressed Kallsyms"))
            self.is_uncompressed = True
            self.markers = self.find_markers_uncompressed()
        else:
            print(message.info("Detected Compressed Kallsyms"))
            self.token_index = self.find_token_index()
            self.markers = self.find_markers()

        self.num_syms = self.find_num_syms()
        self.offsets = self.find_offsets()

        if self.is_offsets:
            self.rbase_offset = self.find_relative_base()

        self.names = self.find_names()
        self.kernel_addresses = self.get_kernel_addresses()
        self.parse_symbol_table()
        for sym_name, sym_addr, sym_type in pwndbg.aglib.kernel.kmod.all_modules_kallsyms():
            self.kallsyms.append((sym_name, sym_type, sym_addr))
        print(message.info(f"Found {len(self.kallsyms)} ksymbols"))

    def find_token_table(self) -> int:
        """
        This function searches for the kallsyms_token_table structure in the kernel memory.
        The kallsyms_token_table contains 256 zero-terminated tokens from which symbol names are built.
        Example structure:
        0xffffffff827b2f00:	"mm"
        0xffffffff827b2f03:	"tim"
        0xffffffff827b2f07:	"bu"
        0xffffffff827b2f0a:	"ode_"
        0xffffffff827b2f0f:	"robestub"
        <SKIPPED>
        0xffffffff827b2fdb:	"0"
        0xffffffff827b2fdd:	"1"
        0xffffffff827b2fdf:	"2"
        0xffffffff827b2fe1:	"3"
        0xffffffff827b2fe3:	"4"
        0xffffffff827b2fe5:	"5"
        0xffffffff827b2fe7:	"6"
        0xffffffff827b2fe9:	"7"
        0xffffffff827b2feb:	"8"
        0xffffffff827b2fed:	"9"
        """
        sequence_to_find = b"".join(b"%c\0" % i for i in range(ord("0"), ord("9") + 1))
        sequences_to_avoid = (b":\0", b"\0\0", b"\0\1", b"\0\2", b"ASCII\0")

        position = 0

        candidates = []
        ascii_candidates = []

        while True:
            position = self.kernel_ro_mem.find(sequence_to_find, position + 1)
            if position == -1:
                break

            for seq in sequences_to_avoid:
                pos = position + len(sequence_to_find)
                if self.kernel_ro_mem[pos : pos + len(seq)] == seq:
                    break
            else:
                candidates.append(position)

                if 32 <= self.kernel_ro_mem[pos : pos + 1][0] < 126:
                    ascii_candidates.append(position)

        if len(candidates) != 1:
            if len(ascii_candidates) == 1:
                candidates = ascii_candidates
            elif len(candidates) == 0:
                # print(M.error("No candidates for token_table, maybe uncompressed kallsyms"))
                return None

        position = candidates[0]

        current_index = 0x30

        position -= 1
        for tokens_backwards in range(current_index):
            for chars_in_token in range(50):
                position -= 1
                assert position >= 0

                if self.kernel_ro_mem[position] == 0 or self.kernel_ro_mem[position] > ord("z"):
                    break

                if chars_in_token >= 50 - 1:
                    print(message.error("This structure is not a kallsyms_token_table"))
                    return None

        position += 1
        position += -position % 4

        return position

    def find_token_index(self) -> int | None:
        """
        This function searches for the kallsyms_token_index structure in the kernel memory
        starting at kallsyms_token_table. The token index table provides offsets into the kallsyms_token_table
        for each 256 byte-valued sub-table.
        The kallsyms_token_index is typically located immediately after
        the kallsyms_token_table in the kernel's read-only data section.
        Example structure:
        0xffffffff827b3288:	0x0000	0x0003	0x0007	0x000a	0x000f	0x0018	0x001f	0x0023
        0xffffffff827b3298:	0x0027	0x0031	0x0035	0x0038	0x003b	0x0043	0x0047	0x004a
        0xffffffff827b32a8:	0x004f	0x0053	0x0056	0x0059	0x005d	0x0061	0x0067	0x006b
        0xffffffff827b32b8:	0x006e	0x0071	0x0076	0x007c	0x0080	0x0088	0x008b	0x008f
        0xffffffff827b32c8:	0x0094	0x0098	0x009b	0x009f	0x00a3	0x00a8	0x00ab	0x00b0
        """
        position = self.token_table

        token_table_head = self.kernel_ro_mem[position : position + 256]

        token_offsets = [p16(0)]

        pos = 0

        while True:
            pos = token_table_head.find(b"\0", pos + 1)
            if pos == -1:
                break
            token_offsets.append(p16(pos + 1))

        seq_to_find = b"".join(token_offsets)

        position = self.kernel_ro_mem.find(seq_to_find, self.token_table)
        if position == -1:
            print(message.error("Unable to find the kallsyms_token_index"))
            return None

        return position

    def find_markers(self) -> int | None:
        """
        This function searches for the kallsyms_markers structure in the kernel memory
        starting at kallsyms_token_table and search backwards. The markers table contains
        offsets to the corresponding symbol name for each kernel symbol.
        The kallsyms_markers table is typically located immediately before the kallsyms_token_table
        in the kernel's read-only data section.
        Example structure:
        0xffffffff827b2430:	0x00000000	0x00000b2a	0x00001762	0x000023f6
        0xffffffff827b2440:	0x00002fe4	0x00003c9d	0x0000487c	0x000056fd
        0xffffffff827b2450:	0x00006597	0x000073b9	0x000081be	0x00008f21
        0xffffffff827b2460:	0x00009c94	0x0000a958	0x0000b632	0x0000c193
        0xffffffff827b2470:	0x0000ce0b	0x0000db98	0x0000ea3e	0x0000f80a
        0xffffffff827b2480:	0x000105be	0x000112d3	0x00011f8c	0x00012d75
        0xffffffff827b2490:	0x0001384d	0x0001446e	0x00015138	0x00015d8c
        """
        if self.kernel_version < (4, 20):
            elem_size = pwndbg.aglib.arch.ptrsize
        else:
            elem_size = 4

        seq_to_find = b"\0" * elem_size

        position = self.token_table - 1

        while position > 0 and self.kernel_ro_mem[position] == 0:
            position -= 1

        for _ in range(32):
            position = self.kernel_ro_mem.rfind(seq_to_find, 0, position)

            if position == -1:
                print(message.error("Failed to find kallsyms_markers"))
                return None

            position -= position % elem_size  # aligning
            size_marker = {4: "I", 8: "Q"}[elem_size]

            entries = unpack_from(f"<4{size_marker}", self.kernel_ro_mem, position)

            if entries[0] != 0:
                continue

            for i in range(1, len(entries)):
                if entries[i - 1] + 0x200 > entries[i] or entries[i - 1] + 0x4000 < entries[i]:
                    break
            else:
                return position

        return None

    def find_num_syms(self):
        """
        This function searches for the kallsyms_num_syms variable in the kernel memory
        starting at kallsyms_markers. The kallsyms_num_syms holds the number of kernel symbols
        in the symbol table.
        The kallsyms_num_syms variable is typically located before the kallsyms_names table in the kernel's
        read-only data section.
        In newer kernel versions the kallsyms_num_syms is immediately behind the linux_banner and in older version
        its behind kallsyms_base_relative or kallsyms_addresses (it depends on CONFIG_KALLSYMS_BASE_RELATIVE y/n)
        """
        if self.kernel_version < (6, 4):
            # try to find num_syms by walking backwards and looking
            # for data like this: a kernel address followed by num_syms
            # 0xffffffff823f8000	0x000000000001417c
            position = self.markers - 8

            while True:
                qword = u64(self.kernel_ro_mem[position : position + 8])
                if (qword >> 32) & 0xFFFFFFFF == 0 and qword > 0:
                    before_qword = u64(self.kernel_ro_mem[position - 8 : position])
                    if (before_qword >> 48) & 0xFFFF == 0xFFFF and (before_qword & 0xFFF) == 0:
                        # should be kallsyms_num_syms
                        return position

                position -= 8
        else:
            # search from kallsyms_markers backwards and look for the linux_banner symbol
            # the kallsyms_num_syms should be behind the linux_banner string
            position = self.kernel_ro_mem.rfind(b"Linux version", 0, self.markers - 8)

            if position == -1:
                return None

            while True:
                position = position + 1
                if self.kernel_ro_mem[position] == 0:
                    break

            position = (position + 7) & ~7  # alignment
            return position

    def find_offsets(self):
        """
        This function searches for the kallsyms_offsets/kallsyms_addresses table in the kernel memory
        starting at kallsyms_token_index. The offsets/addresses table containts offsets / addresses of each
        symbol in the kernel.
        The kallsyms_addresses is typically located before the kallsyms_num_syms variable in the kernel's read-only
        data section.
        Example structure:
        0xffffffff827b3488:	0x00000000	0x00000000	0x00001000	0x00002000
        0xffffffff827b3498:	0x00006000	0x0000b000	0x0000c000	0x0000d000
        0xffffffff827b34a8:	0x00015000	0x00015008	0x00015010	0x00015018
        0xffffffff827b34b8:	0x00015020	0x00015022	0x00015030	0x00015050
        0xffffffff827b34c8:	0x00015450	0x00015460	0x00015860	0x00015888
        0xffffffff827b34d8:	0x00015890	0x00015898	0x000158a0	0x000159c0
        """
        forward_search = self.kernel_version >= (6, 4)

        if forward_search and not self.is_uncompressed:
            position = self.token_index
            self.is_offsets = True
        else:
            # kallsyms_offsets is at the top
            position = self.num_syms
            nsyms = u64(self.kernel_ro_mem[position : position + 8])

            if (
                self.kbase - 0x20000
                < u64(self.kernel_ro_mem[position - 8 : position])
                <= self.kbase
            ):
                # it should be kallsyms_offsets
                self.is_offsets = True
                position -= 8

                dword = u32(self.kernel_ro_mem[position - 4 : position])

                if dword == 0x0:
                    position -= 4

                return position - (nsyms * 4)

            return position - (nsyms * 8)

        while True:
            qword = u64(self.kernel_ro_mem[position : position + 8])
            if qword & 0xFFFFFFFF == 0:
                return position

            position += 8

    def find_relative_base(self):
        """
        This function searches for the kallsyms_relative_base variable in the kernel memory.
        The relative base is used to calculate the actual virtual addresses of symbols from
        their offsets in the kallsyms_offsets table.
        The kallsyms_relative_base variable is typically located after the kallsyms_offsets table
        in the kernel's read-only data section.
        """
        position = self.offsets
        nsyms = u64(self.kernel_ro_mem[self.num_syms : self.num_syms + 8])

        position = position + (nsyms * 4)
        position = (position + 7) & ~7

        return position

    def find_names(self):
        return self.num_syms + 8

    def get_kernel_addresses(self):
        kernel_addresses = []

        rbase = u64(self.kernel_ro_mem[self.rbase_offset : self.rbase_offset + 8])
        # TODO: nsyms is 4 bytes long not 8
        nsyms = u64(self.kernel_ro_mem[self.num_syms : self.num_syms + 8])
        size_marker = "i" if self.is_offsets else "Q"
        kernel_addresses = list(
            unpack_from(f"<{nsyms}{size_marker}", self.kernel_ro_mem, self.offsets)
        )

        if not self.is_offsets:
            return kernel_addresses

        number_of_negative_items = len([offset for offset in kernel_addresses if offset < 0])
        abs_percpu = (len(kernel_addresses) == 0) or (
            number_of_negative_items / len(kernel_addresses) >= 0.5
        )

        for idx, offset in enumerate(kernel_addresses):
            if abs_percpu:
                if offset < 0:
                    offset = rbase - 1 - offset
                else:
                    offset = rbase + offset
            else:
                offset = rbase + offset

            kernel_addresses[idx] = offset

        return kernel_addresses

    def parse_symbol_table(self):
        tokens = self.get_token_table()
        symbol_names = []
        position = self.names
        numsyms = u64(self.kernel_ro_mem[self.num_syms : self.num_syms + 8])

        for _ in range(numsyms):
            length = self.kernel_ro_mem[position]
            position += 1

            symbol_name = ""
            for _ in range(length):
                symbol_token_index = self.kernel_ro_mem[position]
                symbol_token = tokens[symbol_token_index]
                position += 1
                symbol_name += symbol_token
            symbol_names.append(symbol_name)

        for addr, name in zip(self.kernel_addresses, symbol_names):
            self.kallsyms.append((name[1:], name[0], addr))

    def get_token_table(self):
        if not self.is_uncompressed:
            tokens = []
            position = self.token_table

            for num_token in range(256):
                token = ""
                while self.kernel_ro_mem[position]:
                    token += chr(self.kernel_ro_mem[position])
                    position += 1
                position += 1
                tokens.append(token)

        else:
            tokens = [chr(i) for i in range(256)]

        return tokens

    def find_names_uncompressed(self):
        # Find the length byte-separated symbol names
        ksymtab_match = search(
            rb"(?:[\x05-\x23][TWtbBrRAdD][a-z0-9_.]{4,34}){14}", self.kernel_ro_mem
        )

        if not ksymtab_match:
            print(message.error("Failed to find kallsyms"))
            return None

        kallsyms_names__offset = ksymtab_match.start(0)

        # Count the number of symbol names
        position = kallsyms_names__offset
        num_syms = 0

        while position + 1 < len(self.kernel_ro_mem):
            if (
                self.kernel_ro_mem[position] < 2
                or chr(self.kernel_ro_mem[position + 1]).lower() not in "abdrtvwginpcsu-?"
            ):
                break

            symbol_name_and_type = self.kernel_ro_mem[
                position + 1 : position + 1 + self.kernel_ro_mem[position]
            ]

            if not match(rb"^[\x21-\x7e]+$", symbol_name_and_type):
                break

            position += 1 + self.kernel_ro_mem[position]
            num_syms += 1

        if num_syms < 100:
            print(message.error("Failed to find kallsyms"))
            return None

        self.end_of_kallsyms_names_uncompressed = position
        return True

    def find_markers_uncompressed(self):
        """
        This function searches for the kallsyms_markers structure in the kernel memory
        Original Source: https://github.com/marin-m/vmlinux-to-elf/blob/master/vmlinux_to_elf/kallsyms_finder.py
        """
        position = self.end_of_kallsyms_names_uncompressed
        position += -position % 4

        max_number_of_space_between_two_nulls = 0

        # Go just after the first chunk of non-null bytes

        # while position + 1 < len(self.kernel_img) and self.kernel_img[position + 1] == 0:
        #     position += 1

        while position + 1 < len(self.kernel_ro_mem) and self.kernel_ro_mem[position + 1] == 0:
            position += 1

        for null_separated_bytes_chunks in range(20):
            num_non_null_bytes = 1  # we always start at a non-null byte in this loop
            num_null_bytes = (
                1  # we will at least encounter one null byte before the end of this loop
            )

            while True:
                position += 1
                assert position >= 0

                if self.kernel_ro_mem[position] == 0:
                    break
                num_non_null_bytes += 1

            while True:
                position += 1
                assert position >= 0

                if self.kernel_ro_mem[position] != 0:
                    break
                num_null_bytes += 1

            max_number_of_space_between_two_nulls = max(
                max_number_of_space_between_two_nulls, num_non_null_bytes + num_null_bytes
            )

        if (
            max_number_of_space_between_two_nulls % 2 == 1
        ):  # There may be a leap to a shorter offset in the latest processed entries
            max_number_of_space_between_two_nulls -= 1

        if max_number_of_space_between_two_nulls not in (2, 4, 8):
            print(message.error("Could not guess the architecture register size for kernel"))
            return None

        self.offset_table_element_size = max_number_of_space_between_two_nulls

        # Once the size of a long has been guessed, use it to find
        # the first offset (0)

        position = self.end_of_kallsyms_names_uncompressed
        position += -position % 4

        # Go just at the first non-null byte
        while position < len(self.kernel_ro_mem) and self.kernel_ro_mem[position] == 0:
            position += 1

        likely_is_big_endian = position % self.offset_table_element_size > 1
        if self.is_big_endian is None:  # Manual architecture specification
            self.is_big_endian = likely_is_big_endian

        if position % self.offset_table_element_size == 0:
            position += self.offset_table_element_size
        else:
            position += -position + self.offset_table_element_size

        position -= self.offset_table_element_size
        position -= self.offset_table_element_size

        position -= position % self.offset_table_element_size

        self.kallsyms_markers__offset = position

        # print('Found kallsyms_markers at file offset 0x%08x' % position)

        return position



"""
Dynamic configuration system for pwndbg, using GDB's built-in Parameter
mechanism.

To create a new pwndbg configuration point, call ``pwndbg.config.add_param``.

Parameters should be declared in the module in which they are primarily
used, or in this module for general-purpose parameters.

All pwndbg Parameter types are accessible via property access on this
module, for example:

    >>> pwndbg.config.add_param('example-value', 7, 'an example')
    >>> int(pwndbg.config.example_value)
    7
"""

from __future__ import annotations

from typing import Any

import gdb

import pwndbg
import pwndbg.decorators
import pwndbg.lib.config

CLASS_MAPPING = {
    pwndbg.lib.config.PARAM_BOOLEAN: gdb.PARAM_BOOLEAN,
    pwndbg.lib.config.PARAM_AUTO_BOOLEAN: gdb.PARAM_AUTO_BOOLEAN,
    pwndbg.lib.config.PARAM_INTEGER: gdb.PARAM_INTEGER,
    pwndbg.lib.config.PARAM_ZINTEGER: gdb.PARAM_ZINTEGER,
    pwndbg.lib.config.PARAM_UINTEGER: gdb.PARAM_UINTEGER,
    pwndbg.lib.config.PARAM_ZUINTEGER: gdb.PARAM_ZUINTEGER,
    pwndbg.lib.config.PARAM_ZUINTEGER_UNLIMITED: gdb.PARAM_ZUINTEGER_UNLIMITED,
    pwndbg.lib.config.PARAM_STRING: gdb.PARAM_STRING,
    pwndbg.lib.config.PARAM_ENUM: gdb.PARAM_ENUM,
    pwndbg.lib.config.PARAM_OPTIONAL_FILENAME: gdb.PARAM_OPTIONAL_FILENAME,
}


# See this for details about the API of `gdb.Parameter`:
# https://sourceware.org/gdb/onlinedocs/gdb/Parameters-In-Python.html
class Parameter(gdb.Parameter):
    def __init__(self, param: pwndbg.lib.config.Parameter) -> None:
        # `set_doc`, `show_doc`, and `__doc__` must be set before `gdb.Parameter.__init__`.
        # They will be used for `help set <param>` and `help show <param>`,
        # respectively
        self.set_doc = "Set " + param.set_show_doc + "."
        self.show_doc = "Show " + param.set_show_doc + "."
        self.__doc__ = param.help_docstring or None

        self.init_super(param)
        self.param = param
        self.value = param.value
        self.param.add_update_listener(self.on_change)

    def init_super(self, param: pwndbg.lib.config.Parameter) -> None:
        """Initializes the super class for GDB >= 9"""
        c = CLASS_MAPPING[param.param_class]
        if c == gdb.PARAM_ENUM:
            super().__init__(
                param.name,
                gdb.COMMAND_SUPPORT,
                c,
                param.enum_sequence,
            )
            return
        super().__init__(param.name, gdb.COMMAND_SUPPORT, c)

    def on_change(self, value: Any) -> None:
        """Called when the value of the pwndbg.lib.config.Parameter changes
        Transfer the value to the GDB parameter to keep them in sync.
        """
        self.value = value

    @property
    def native_value(self):
        return Parameter._value_to_gdb_native(
            self.param.value, gdb_param_class=CLASS_MAPPING[self.param.param_class]
        )

    @property
    def native_default(self):
        return Parameter._value_to_gdb_native(
            self.param.default, gdb_param_class=CLASS_MAPPING[self.param.param_class]
        )

    def get_set_string(self) -> str:
        """Handles the GDB `set <param>`"""
        # GDB will set `self.value` to the user's input
        if self.value is None and CLASS_MAPPING[self.param.param_class] in (
            gdb.PARAM_UINTEGER,
            gdb.PARAM_INTEGER,
        ):
            # FIXME: The comment below is wrong, see
            # https://sourceware.org/gdb/current/onlinedocs/gdb.html/Parameters-In-Python.html#Parameters-In-Python
            # Do we need this branch?
            # # Note: This is really weird, according to GDB docs, 0 should mean "unlimited" for gdb.PARAM_UINTEGER and gdb.PARAM_INTEGER, but somehow GDB sets the value to `None` actually :/
            # # And hilarious thing is that GDB won't let you set the default value to `None` when you construct the `gdb.Parameter` object with `gdb.PARAM_UINTEGER` or `gdb.PARAM_INTEGER` lol
            # # Maybe it's a bug of GDB?
            # # Anyway, to avoid some unexpected behaviors, we'll still set `self.param.value` to 0 here.
            self.param.value = 0
        else:
            self.param.value = self.value

        for trigger in pwndbg.config.triggers[self.param.name]:
            trigger()

        # No need to print anything if this is set before we get to a prompt,
        # like if we're setting options in .gdbinit
        if not pwndbg.decorators.first_prompt:
            return ""

        return f"Set {self.param.set_show_doc} to {self.native_value!r}."

    def get_show_string(self, svalue: str) -> str:
        """Handles the GDB `show <param>`"""
        more_information_hint = f" See `help set {self.param.name}` for more information."
        return "{} is {!r}.{}".format(
            self.param.set_show_doc.capitalize(),
            svalue,
            more_information_hint if self.__doc__ else "",
        )

    @staticmethod
    def _value_to_gdb_native(value: Any, gdb_param_class: int | None = None) -> Any:
        """Translates Python value into native GDB syntax string."""
        if isinstance(value, bool):
            # Convert booleans to "on" or "off".
            return "on" if value else "off"
        if value is None and gdb_param_class == gdb.PARAM_AUTO_BOOLEAN:
            # None for gdb.PARAM_AUTO_BOOLEAN means "auto".
            return "auto"
        if value == 0 and gdb_param_class in (gdb.PARAM_UINTEGER, gdb.PARAM_INTEGER):
            # 0 for gdb.PARAM_UINTEGER and gdb.PARAM_INTEGER means "unlimited".
            return "unlimited"
        if value == -1 and gdb_param_class == gdb.PARAM_ZUINTEGER_UNLIMITED:
            # -1 for gdb.PARAM_ZUINTEGER_UNLIMITED means "unlimited".
            return "unlimited"

        # Other types pass through normally
        return value


def init_params() -> None:
    # Create a gdb.Parameter for each parameter
    for p in pwndbg.config.params.values():
        # We don't need to store this anywhere, GDB will handle this
        Parameter(p)



from __future__ import annotations

import argparse

import pwndbg.aglib.proc
import pwndbg.aglib.vmmap
import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory

if pwndbg.dbg.is_gdblib_available():
    import gdb


def translate_addr(offset, module):
    mod_filter = lambda page: module in page.objfile
    pages = list(filter(mod_filter, pwndbg.aglib.vmmap.get()))

    if not pages:
        print(
            "There are no memory pages in `vmmap` "
            f"for specified address=0x{offset:x} and module={module}"
        )
        return

    first_page = min(pages, key=lambda page: page.vaddr)

    addr = first_page.vaddr + offset

    if not any(addr in p for p in pages):
        print(
            f"Offset 0x{offset:x} rebased to module {module} as 0x{addr:x} is beyond module's "
            "memory pages:"
        )
        for p in pages:
            print(p)
        return

    return addr


parser = argparse.ArgumentParser(description="Calculate VA of RVA from PIE base.")
parser.add_argument("offset", nargs="?", type=int, default=0, help="Offset from PIE base.")
parser.add_argument(
    "module",
    type=str,
    nargs="?",
    default="",
    help="Module to choose as base. Defaults to the target executable.",
)


@pwndbg.commands.Command(parser, category=CommandCategory.LINUX)
@pwndbg.commands.OnlyWhenRunning
def piebase(offset: int = 0, module: str = "") -> None:
    offset = int(offset)
    if not module:
        module = pwndbg.aglib.proc.exe()

    addr = translate_addr(offset, module)

    if addr is not None:
        print(f"Calculated VA from {module} = 0x{addr:x}")
    else:
        print(message.error("Could not calculate VA on current target."))


if pwndbg.dbg.is_gdblib_available():
    parser = argparse.ArgumentParser()
    parser.description = "Break at RVA from PIE base."
    parser.add_argument("offset", nargs="?", type=int, default=0, help="Offset to add.")
    parser.add_argument(
        "module",
        type=str,
        nargs="?",
        default="",
        help="Module to choose as base. Defaults to the target executable.",
    )

    @pwndbg.commands.Command(parser, aliases=["brva"], category=CommandCategory.BREAKPOINT)
    @pwndbg.commands.OnlyWhenRunning
    def breakrva(offset: int = 0, module: str = "") -> None:
        offset = int(offset)
        if not module:
            module = pwndbg.aglib.proc.exe()

        addr = translate_addr(offset, module)

        if addr is not None:
            spec = f"*{addr:#x}"
            gdb.Breakpoint(spec)
        else:
            print(message.error("Could not determine rebased breakpoint address on current target"))



from __future__ import annotations

import re
from typing import Any

import pwndbg.aglib.disasm.disassembly
import pwndbg.aglib.kernel
import pwndbg.aglib.memory
import pwndbg.aglib.qemu
import pwndbg.aglib.symbol
import pwndbg.dbg_mod
import pwndbg.lib.cache
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.dbg_mod import EventType
from pwndbg.lib import TypeNotRecoveredError

#########################################
# helpers
#
#########################################
POSSIBLE_ZONE_NAMES = (
    "DMA",
    "DMA32",
    "Normal",
    "HighMem",
    "Movable",
    "Device",
)


@pwndbg.lib.cache.cache_until("objfile")
def migratetype_names() -> tuple[str, ...]:
    names = [
        "Unmovable",
        "Movable",
        "Reclaimable",
        "HighAtomic",
    ]
    kconfig = pwndbg.aglib.kernel.kconfig()
    if "CONFIG_CMA" in kconfig:
        names.append("CMA")
    if "CONFIG_MEMORY_ISOLATION" in kconfig:
        names.append("Isolate")
    return tuple(names)


# try getting value of a symbol as an unsigned integer
def try_usymbol(name: str, size: int | None = None) -> int | None:
    if not pwndbg.aglib.kernel.has_debug_symbols():
        return None
    try:
        if pwndbg.aglib.kernel.has_debug_info():
            return pwndbg.aglib.symbol.lookup_symbol_value(name)

        symbol = pwndbg.aglib.symbol.lookup_symbol_addr(name)
        if symbol is None:
            return None

        if size is None:
            size = pwndbg.aglib.arch.ptrbits

        if size == 8:
            return pwndbg.aglib.memory.u(symbol)
        if size == 16:
            return pwndbg.aglib.memory.u16(symbol)
        if size == 32:
            return pwndbg.aglib.memory.u32(symbol)

        return pwndbg.aglib.memory.u64(symbol)
    except Exception:
        # for kpti
        return None


@pwndbg.aglib.kernel.requires_debug_symbols("zone_names", default=4)
def nzones() -> int:
    _zone_names = pwndbg.aglib.symbol.lookup_symbol_addr("zone_names")
    for i in range(len(POSSIBLE_ZONE_NAMES) + 1):
        char_ptr = pwndbg.aglib.memory.u64(_zone_names + i * 8)
        if pwndbg.aglib.memory.string(char_ptr).decode() not in POSSIBLE_ZONE_NAMES:
            return i
    assert False, "cannot determine the number of zones"


def nmtypes() -> int:
    return len(migratetype_names())


def npcplist() -> int:
    """returns NR_PCP_LISTS (https://elixir.bootlin.com/linux/v6.13/source/include/linux/mmzone.h#L671)"""
    if (
        not pwndbg.aglib.kernel.has_debug_symbols("node_zones")
        or not pwndbg.aglib.kernel.has_debug_info()
    ):
        if pwndbg.aglib.kernel.krelease() < (5, 14):
            return 3
        return 12
    node_data0 = pwndbg.aglib.kernel.node_data()
    if "CONFIG_NUMA" in pwndbg.aglib.kernel.kconfig():
        node_data0 = node_data0.dereference()
    zone = node_data0[0]["node_zones"][0]
    # index 0 should always exist
    if zone.type.has_field("per_cpu_pageset"):
        lists = zone["per_cpu_pageset"]["lists"]
        return lists.type.array_len
    if zone.type.has_field("pageset"):
        lists = zone["pageset"]["pcp"]["lists"]
        return lists.type.array_len
    return 0


def kversion_cint(kversion: tuple[int, ...] | None = None) -> int | None:
    if kversion is None:
        kversion = pwndbg.aglib.kernel.krelease()
    if kversion is None or len(kversion) != 3:
        return None
    x, y, z = kversion
    return ((x) * 65536) + ((y) * 256) + (z)


#########################################
# common structurs
#
#########################################
COMMON_TYPES = """
#include <stdint.h>
#include <stddef.h>
#include <linux/version.h>
typedef unsigned char u8;
typedef char s8;
typedef unsigned short u16;
typedef unsigned int u32;
typedef long long s64;
typedef unsigned long u64;
#define bool int
#if UINTPTR_MAX == 0xffffffff
    typedef int16_t arch_word_t;
#else
    typedef int32_t arch_word_t;
#endif
typedef struct {
    unsigned int val;
} kuid_t;
typedef struct {
    unsigned int val;
} kgid_t;
typedef int pid_t;
typedef struct {
    int counter;
} atomic_t;
typedef struct refcount_struct {
	atomic_t refs;
} refcount_t;

struct list_head {
    struct list_head *next, *prev;
};
struct hlist_node {
	struct hlist_node *next, **pprev;
};
struct kmem_cache;
enum pageflags {
	PG_locked,		/* Page is locked. Don't touch. */
	PG_writeback,		/* Page is under writeback */
	PG_referenced,
	PG_uptodate,
	PG_dirty,
	PG_lru,
	PG_head,		/* Must be in bit 6 */
	PG_waiters,		/* Page has waiters, check its waitqueue. Must be bit #7 and in the same byte as "PG_locked" */
	PG_active,
	PG_workingset,
	PG_owner_priv_1,	/* Owner use. If pagecache, fs may use */
	PG_owner_2,		/* Owner use. If pagecache, fs may use */
	PG_arch_1,
	PG_reserved,
	PG_private,		/* If pagecache, has fs-private data */
	PG_private_2,		/* If pagecache, has fs aux data */
	PG_reclaim,		/* To be reclaimed asap */
	PG_swapbacked,		/* Page is backed by RAM/swap */
	PG_unevictable,		/* Page is "unevictable"  */
	PG_dropbehind,		/* drop pages on IO completion */
};
#define POINTER_SIZE (sizeof(void *))
"""


@pwndbg.aglib.kernel.typeinfo_recovery("struct page", requires_kversion=True)
def recover_page_typeinfo() -> str:
    defs = []
    for config in (
        "CONFIG_MEMCG",
        "CONFIG_KASAN",
    ):
        if config in pwndbg.aglib.kernel.kconfig():
            defs.append(config)
    result = f"#define KVERSION {kversion_cint()}\n"
    result += "\n".join(f"#define {s}" for s in defs)
    result += COMMON_TYPES
    result += """
    struct page { // just a simplied page struct with relavent fields
        unsigned long flags;
        union {
            struct {
                union {
                    struct {
                        union {
                            struct list_head lru;
                            struct list_head buddy_list;
                            struct list_head pcp_list;
                        };
                    };
                    struct {	/* Tail pages of compound page */
                        unsigned long compound_head;	/* Bit zero is set */
                    };
                };
            };
#if KVERSION < KERNEL_VERSION(5, 17, 0)
            struct {	/* slab, slob and slub */
                union {
                    struct list_head slab_list;
                    struct {	/* Partial pages */
                        struct page *next;
                        arch_word_t pages;	/* Nr of pages left */
#if KVERSION < KERNEL_VERSION(5, 16, 0)
                        arch_word_t pobjects;	/* Approximate count */
#endif
                    };
                };
                struct kmem_cache *slab_cache; /* not slob */
                /* Double-word boundary */
                void *freelist;		/* first free object */
                union {
                    void *s_mem;	/* slab: first object */
                    unsigned long counters;		/* SLUB */
                    struct {			/* SLUB */
                        unsigned inuse:16;
                        unsigned objects:15;
                        unsigned frozen:1;
                    };
                };
            };
#endif
            char _pad[0x28]; // the rest are not relavent to this project but needs to be 0x28 bytes
        };
        union {
            atomic_t _mapcount;
            unsigned int page_type;
        };
        atomic_t _refcount;
#ifdef CONFIG_MEMCG
        unsigned long memcg_data;
#endif
#if defined(WANT_PAGE_VIRTUAL) /* never set for x86 and arm */
        void *virtual;
#endif /* WANT_PAGE_VIRTUAL */
#ifdef LAST_CPUPID_NOT_IN_PAGE_FLAGS /* TODO: seems never got set for all the kernel builds I have worked with */
#if KVERSION >= KERNEL_VERSION(6, 7, 0)
        int _last_cpupid;
#endif
#endif
#if defined(CONFIG_KASAN) && KVERSION >= KERNEL_VERSION(6, 1, 0)
        struct page *kmsan_shadow;
        struct page *kmsan_origin;
#endif
#ifdef LAST_CPUPID_NOT_IN_PAGE_FLAGS
#if KVERSION < KERNEL_VERSION(6, 7, 0)
        int _last_cpupid;
#endif
#endif
    };
    """
    return result


@pwndbg.dbg.event_handler(EventType.NEW_MODULE)
def load_common_structs_on_load_linux() -> None:
    if pwndbg.aglib.qemu.is_qemu_kernel() and pwndbg.dbg.selected_inferior().is_linux():
        try:
            recover_page_typeinfo()
        except TypeNotRecoveredError as e:
            # We are not going to print anything here, because the user may not
            # even end up using the type-dependant commands.
            # Other commands and typeinfo recoveries depend on this succeeding,
            # so we save the actual failure reason to have something meaningful to
            # show to the user.
            pwndbg.aglib.kernel.page_typeinfo_recovery_failure = e


class ArchSymbols:
    def __init__(self) -> None:
        self.node_data_heuristic_func = "first_online_pgdat"
        self.slab_caches_heuristic_func = "slab_next"
        self.per_cpu_offset_heuristic_func = "nr_iowait_cpu"
        self.modules_heuristic_func = "find_module_all"
        self.db_list_heuristic_func = (
            "dma_buf_file_release"
            if pwndbg.aglib.kernel.krelease() >= (5, 10)
            else "dma_buf_release"
        )
        self.bpf_prog_heuristic_func = "bpf_prog_free_id"
        self.bpf_map_heuristic_func = "bpf_map_free_id"
        self.current_task_heuristic_func = "common_cpu_up"

    def disass(self, name: str, lines: int = 10) -> str | None:
        sym = pwndbg.aglib.symbol.lookup_symbol(name)
        if sym is None:
            return None
        addr = int(sym)
        disass = []
        for _ in range(lines):
            instr: PwndbgInstruction = pwndbg.aglib.disasm.disassembly.get_one_instruction(
                addr, enhance=False
            )
            disass.append(hex(addr) + " " + instr.asm_string)
            addr = instr.next
        return "\n".join(disass)

    def regex(self, s: str, pattern: str, nth: int) -> re.Match[Any] | None:
        p = re.compile(pattern)
        if nth == 0:
            return p.search(s)
        matches = list(p.finditer(s))
        if nth < len(matches):
            return matches[nth]
        return None

    @pwndbg.lib.cache.cache_until("stop")
    def node_data(self) -> pwndbg.dbg_mod.Value:
        node_data = pwndbg.aglib.symbol.lookup_symbol("node_data")
        if pwndbg.aglib.kernel.has_debug_info():
            return node_data
        if node_data is None and pwndbg.aglib.kernel.has_debug_symbols(
            self.node_data_heuristic_func
        ):
            node_data = self._node_data()
        return pwndbg.aglib.memory.get_typed_pointer("unsigned long", node_data)

    @pwndbg.lib.cache.cache_until("stop")
    def slab_caches(self) -> pwndbg.dbg_mod.Value:
        slab_caches = pwndbg.aglib.symbol.lookup_symbol("slab_caches")
        if slab_caches is None and pwndbg.aglib.kernel.has_debug_symbols(
            self.slab_caches_heuristic_func
        ):
            slab_caches = self._slab_caches()
        return pwndbg.aglib.memory.get_typed_pointer_value("struct list_head", slab_caches)

    @pwndbg.lib.cache.cache_until("stop")
    def per_cpu_offset(self) -> pwndbg.dbg_mod.Value:
        per_cpu_offset = pwndbg.aglib.symbol.lookup_symbol("__per_cpu_offset")
        if per_cpu_offset is not None:
            return per_cpu_offset
        if pwndbg.aglib.kernel.has_debug_symbols(self.per_cpu_offset_heuristic_func):
            per_cpu_offset = self._per_cpu_offset()
        return pwndbg.aglib.memory.get_typed_pointer("unsigned long", per_cpu_offset)

    @pwndbg.lib.cache.cache_until("stop")
    def modules(self) -> pwndbg.dbg_mod.Value:
        modules = pwndbg.aglib.symbol.lookup_symbol("modules")
        if modules:
            return modules
        if pwndbg.aglib.kernel.has_debug_symbols(self.modules_heuristic_func):
            modules = self._modules()
        return pwndbg.aglib.memory.get_typed_pointer("unsigned long", modules)

    @pwndbg.lib.cache.cache_until("stop")
    def db_list(self) -> pwndbg.dbg_mod.Value:
        if pwndbg.aglib.kernel.krelease() >= (6, 10):
            debugfs_list = pwndbg.aglib.symbol.lookup_symbol("debugfs_list")
            # TODO: fallback not supported for >= v6.10, should look at dma_buf_debug_show later if needed
            # though the symbol should exist if the function symbol exist
            return debugfs_list
        db_list = pwndbg.aglib.symbol.lookup_symbol("db_list")
        if db_list:
            return db_list
        if pwndbg.aglib.kernel.has_debug_symbols(self.db_list_heuristic_func):
            db_list = self._db_list()
        return pwndbg.aglib.memory.get_typed_pointer("struct list_head", db_list)

    @pwndbg.lib.cache.cache_until("stop")
    def map_idr(self) -> pwndbg.dbg_mod.Value:
        map_idr = pwndbg.aglib.symbol.lookup_symbol("map_idr")
        if map_idr:
            return map_idr
        if pwndbg.aglib.kernel.has_debug_symbols(self.bpf_map_heuristic_func):
            map_idr = self._map_idr()
        return pwndbg.aglib.memory.get_typed_pointer("unsigned long", map_idr)

    @pwndbg.lib.cache.cache_until("stop")
    def prog_idr(self) -> pwndbg.dbg_mod.Value:
        prog_idr = pwndbg.aglib.symbol.lookup_symbol("prog_idr")
        if prog_idr:
            return prog_idr
        if pwndbg.aglib.kernel.has_debug_symbols(self.bpf_prog_heuristic_func):
            prog_idr = self._prog_idr()
        return pwndbg.aglib.memory.get_typed_pointer("unsigned long", prog_idr)

    @pwndbg.lib.cache.cache_until("stop")
    def current_task(self, cpu: int | None) -> int | None:
        # using symbols usually yield incorrect results
        current_task = None
        if pwndbg.aglib.arch.name == "aarch64":
            current_task = self._current_task()
        elif pwndbg.aglib.kernel.has_debug_symbols(self.current_task_heuristic_func):
            current_task = self._current_task()
            if current_task is not None:
                current_task = pwndbg.aglib.kernel.per_cpu(current_task, cpu=cpu)
                # current_task is int but needed here to make the linter happy
                current_task = pwndbg.aglib.memory.read_pointer_width(int(current_task))
        return current_task

    def _node_data(self) -> int | None:
        raise NotImplementedError()

    def _slab_caches(self) -> int | None:
        raise NotImplementedError()

    def _per_cpu_offset(self) -> int | None:
        raise NotImplementedError()

    def _modules(self) -> int | None:
        raise NotImplementedError()

    def _db_list(self) -> int | None:
        raise NotImplementedError()

    def _map_idr(self) -> int | None:
        raise NotImplementedError()

    def _prog_idr(self) -> int | None:
        raise NotImplementedError()

    def _current_task(self) -> int | None:
        raise NotImplementedError()


class x86_64Symbols(ArchSymbols):
    # op ... [... +/- (0x...)]
    # if negative, the `-0x...`` is a kernel address displayed as a negative number
    # returns the first 0x... as an int if exists
    def qword_op_reg_memoff(
        self, disass: str, op: str, sign: str = "-", nth: int = 0
    ) -> int | None:
        result = self.regex(disass, rf"{op}.*\[.*{re.escape(sign)}\s(0x[0-9a-f]+)\]", nth)
        if result is not None:
            if sign == "-":
                return (1 << 64) - int(result.group(1), 16)
            return int(result.group(1), 16)
        return None

    # mov reg, <kernel address as a constant>
    def qword_mov_reg_const(self, disass: str, nth: int = 0) -> int | None:
        result = self.regex(disass, r"mov.*(0x[0-9a-f]{16})", nth)
        if result is not None:
            return int(result.group(1), 16)
        return None

    def dword_mov_reg_const(self, disass: str, nth: int = 0) -> int | None:
        result = self.regex(disass, r"mov.*(0x[0-9a-f]{1,8})\b(?!\])", nth)
        if result is not None:
            return int(result.group(1), 16)
        return None

    def qword_mov_reg_ripoff(self, disass: str, nth: int = 0) -> int | None:
        result = self.regex(
            " ".join(disass.splitlines()),
            r"mov.*\[rip\s([\+\-]\s0x[0-9a-f]+)\]\s(0x[0-9a-f]{16})",
            nth,
        )
        if result is not None:
            return int(result.group(1).replace(" ", ""), 16) + int(result.group(2), 16)
        return None

    def _node_data(self) -> int | None:
        disass = self.disass(self.node_data_heuristic_func)
        if not disass:
            return None
        result = self.qword_op_reg_memoff(disass, op="mov", sign="-")
        if result is not None:
            return result
        return self.qword_mov_reg_const(disass)

    def _slab_caches(self) -> int | None:
        disass = self.disass(self.slab_caches_heuristic_func)
        if not disass:
            return None
        return self.qword_mov_reg_const(disass)

    def _per_cpu_offset(self) -> int | None:
        disass = self.disass(self.per_cpu_offset_heuristic_func)
        if not disass:
            return None
        result = self.qword_op_reg_memoff(disass, op="add", sign="-")
        if result is not None:
            return result
        result = self.qword_mov_reg_const(disass)
        if result is not None:
            return result
        return self.qword_mov_reg_ripoff(disass)

    def _modules(self) -> int | None:
        disass = self.disass(self.modules_heuristic_func)
        if not disass:
            return None
        return self.qword_mov_reg_ripoff(disass)

    def _db_list(self) -> int | None:
        offset = 0x10  # offset of the lock
        disass = self.disass(self.db_list_heuristic_func)
        if not disass:
            return None
        result = self.qword_mov_reg_const(disass)
        if result is not None:
            return result - offset
        return None

    def _map_idr(self) -> int | None:
        disass = self.disass(self.bpf_map_heuristic_func, lines=50)
        if not disass:
            return None
        result = self.qword_mov_reg_const(disass, nth=1)
        if result is not None:
            return result
        return self.qword_mov_reg_const(disass)

    def _prog_idr(self) -> int | None:
        disass = self.disass(self.bpf_prog_heuristic_func, lines=50)
        if not disass:
            return None
        result = self.qword_mov_reg_const(disass, nth=1)
        if result is not None:
            return result
        return self.qword_mov_reg_const(disass)

    def _current_task(self) -> int | None:
        disass = self.disass(self.current_task_heuristic_func)
        if not disass:
            return None
        result = self.dword_mov_reg_const(disass)
        if result is not None:
            return result
        result = self.qword_mov_reg_const(disass)
        return result


class Aarch64Symbols(ArchSymbols):
    # adrp x?, <kernel address>
    # add x?, x?, #0x...
    def qword_adrp_add_const(self, disass: str, nth: int = 0) -> int | None:
        prev = ""
        for line in disass.splitlines():
            if "adrp" in prev and "add" in line:
                result = self.regex(prev, r"\,\s*0x([0-9a-f]+)", nth=0)
                tmp = None
                if result is not None:
                    tmp = int(result.group(1), 16)
                result = self.regex(line, r"#0x([0-9a-f]+)", nth=0)
                if result is not None and tmp is not None:
                    if nth == 0:
                        return tmp + int(result.group(1), 16)
                    nth -= 1
            prev = line
        return None

    def _node_data(self) -> int | None:
        disass = self.disass(self.node_data_heuristic_func)
        if not disass:
            return None
        return self.qword_adrp_add_const(disass)

    def _slab_caches(self) -> int | None:
        disass = self.disass(self.slab_caches_heuristic_func)
        if not disass:
            return None
        result = self.qword_adrp_add_const(disass)
        if result:
            return result
        # adrp x<num>, 0x....
        # ...
        # add x<num>, x<num>, #0x...
        # ...
        # add x1, x<num>, #0x...
        pattern = re.compile(
            r"adrp\s+x(\d+),\s+0x([0-9a-fA-F]+).*?\n"
            r".*?add\s+x\1,\s+x\1,\s+#0x([0-9a-fA-F]+).*?\n"
            r".*?add\s+x1,\s+x\1,\s+#0x([0-9a-fA-F]+)",
            re.DOTALL,
        )
        m = pattern.search(disass)
        if m is None:
            return None
        return sum(int(m.group(i), 16) for i in [2, 3, 4])

    def _per_cpu_offset(self) -> int | None:
        disass = self.disass(self.per_cpu_offset_heuristic_func)
        if not disass:
            return None
        return self.qword_adrp_add_const(disass)

    def _modules(self) -> int | None:
        disass = self.disass(self.modules_heuristic_func)
        if not disass:
            return None
        # adrp x<num>, 0x....
        # ...
        # add x<num>, x<num>, #0x...
        # ...
        # ldr x?, [x<num>, #0x]!...
        pattern = re.compile(
            r"adrp\s+x(\d+),\s+0x([0-9a-fA-F]+).*?\n"
            r".*?add\s+x\1,\s+x\1,\s+#0x([0-9a-fA-F]+).*?\n"
            r".*?ldr\s+x\d+,\s+\[x\1,\s+#0x([0-9a-fA-F]+)\]!",
            re.DOTALL,
        )
        m = pattern.search(disass)
        if m is None:
            return None
        return sum(int(m.group(i), 16) for i in [2, 3, 4])

    def _db_list(self) -> int | None:
        offset = 0x10  # offset of the lock
        disass = self.disass(self.db_list_heuristic_func)
        if not disass:
            return None
        result = self.qword_adrp_add_const(disass)
        if result is not None:
            return result - offset
        return None

    def _map_idr(self) -> int | None:
        disass = self.disass(self.bpf_map_heuristic_func, lines=50)
        if not disass:
            return None
        result = self.qword_adrp_add_const(disass, nth=1)
        if result is not None:
            return result
        return self.qword_adrp_add_const(disass)

    def _prog_idr(self) -> int | None:
        disass = self.disass(self.bpf_prog_heuristic_func, lines=50)
        if not disass:
            return None
        result = self.qword_adrp_add_const(disass, nth=1)
        if result is not None:
            return result
        return self.qword_adrp_add_const(disass)

    def _current_task(self) -> int | None:
        return pwndbg.aglib.regs.read_reg("sp_el0")
